#!/usr/bin/env python
'''
SymbolTable for Larch interpreter
'''
import copy
import numpy
from lmfit.printfuncs import gformat
from . import site_config
from .utils import fixName, isValidName

def repr_value(val):
    """render a repr-like value for ndarrays, lists, etc"""
    if (isinstance(val, numpy.ndarray) and
            (len(val) > 6 or len(val.shape)>1)):
        sval = f"shape={val.shape}, type={val.dtype} range=[{gformat(val.min())}:{gformat(val.max())}]"
    elif isinstance(val, list) and len(val) > 6:
        sval = f"length={len(val)}: [{val[0]}, {val[1]}, ... {val[-2]}, {val[-1]}]"
    elif isinstance(val, tuple) and len(val) > 6:
        sval = f"length={len(val)}: ({val[0]}, {val[1]}, ... {val[-2]}, {val[-1]})"
    else:
        try:
            sval = repr(val)
        except:
            sval = val
    return sval


class Group():
    """
    Generic Group: a container for variables, modules, and subgroups.
    """
    __private = ('_main', '_larch', '_parents', '__name__', '__doc__',
                 '__private', '_subgroups', '_members', '_repr_html_')

    __generic_functions = ('keys', 'values', 'items')

    def __init__(self, name=None, **kws):
        if name is None:
            name = hex(id(self))
        self.__name__ = name
        for key, val in kws.items():
            setattr(self, key, val)

    def __len__(self):
        return len(dir(self))

    def __repr__(self):
        if self.__name__ is not None:
            return f'<Group {self.__name__}>'
        return '<Group>'

    def __copy__(self):
        out = Group()
        for key, val in self.__dict__.items():
            if key != '__name__':
                setattr(out, key,  copy.copy(val))
        return out

    def __deepcopy__(self, memo):
        out = Group()
        for key, val in self.__dict__.items():
            if key != '__name__':
                setattr(out, key,  copy.deepcopy(val, memo))
        return out

    def __id__(self):
        return id(self)

    def __dir__(self):
        "return list of member names"
        cls_members = []
        cname = self.__class__.__name__
        if cname != 'SymbolTable' and hasattr(self, '__class__'):
            cls_members = dir(self.__class__)

        dict_keys = [key for key in self.__dict__ if key not in cls_members]

        return [key for key in cls_members + dict_keys
                if (not key.startswith('_SymbolTable_') and
                    not key.startswith('_Group_') and
                    not key.startswith(f'_{cname}_') and
                    not (key.startswith('__') and key.endswith('__')) and
                    key not in self.__generic_functions and
                    key not in self.__private)]

    def __getitem__(self, key):

        if isinstance(key, int):
            raise IndexError("Group does not support Integer indexing")

        return getattr(self, key)

    def __setitem__(self, key, value):

        if isinstance(key, int):
            raise IndexError("Group does not support Integer indexing")

        return setattr(self, key, value)

    def __iter__(self):
        return iter(self.keys())

    def keys(self):
        return self.__dir__()

    def values(self):
        return [getattr(self, key) for key in self.__dir__()]

    def items(self):
        return [(key, getattr(self, key)) for key in self.__dir__()]

    def _subgroups(self):
        "return list of names of members that are sub groups"
        return [k for k in self._members() if isgroup(self.__dict__[k])]

    def _members(self):
        "return members"
        out = {}
        for key in self.__dir__():
            if key in self.__dict__:
                out[key] = self.__dict__[key]
        return out

    def _repr_html_(self):
        """HTML representation for Jupyter notebook"""
        html = [f"Group {self.__name__}", "<table>",
                "<tr><td><b>Attribute</b></td><td><b>Type</b></td>",
                "<td><b>Value</b></td></tr>"]
        attrs = self.__dir__()
        for attr in self.__dir__():
            obj = getattr(self, attr)
            atype = type(obj).__name__
            sval = repr_value(obj)
            html.append(f"<tr><td>{attr}</td><td><i>{atype}</i></td><td>{sval}</td></tr>")
        html.append("</table>")
        return '\n'.join(html)


def isgroup(grp, *args):
    """tests if input is a Group

    With additional arguments (all must be strings), it also tests
    that the group has an an attribute named for each argument. This
    can be used to test not only if a object is a Group, but whether
    it a group with expected arguments.
    """
    ret = isinstance(grp, Group)
    if ret and len(args) > 0:
        try:
            ret = all([hasattr(grp, a) for a in args])
        except TypeError:
            return False
    return ret


class InvalidName:
    """ used to create a value that will NEVER be a useful symbol.
    symboltable._lookup() uses this to check for invalid names"""


GroupDocs = {}
GroupDocs['_sys'] = """
Larch system-wide status variables, including
configuration variables and lists of Groups used
for finding variables.
"""

GroupDocs['_builtin'] = """
core built-in functions, most taken from Python
"""

GroupDocs['_math'] = """
Mathematical functions, including a host of functtion from numpy and scipy
"""


class SymbolTable(Group):
    """Main Symbol Table for Larch.
    """
    top_group   = '_main'
    core_groups = ('_sys', '_builtin', '_math')
    __invalid_name = InvalidName()
    _private = ('save_frame', 'restore_frame', 'set_frame',
                'has_symbol', 'has_group', 'get_group',
                'create_group', 'new_group', 'isgroup',
                'get_symbol', 'set_symbol',  'del_symbol',
                'get_parent', '_path', '__parents')

    def __init__(self, larch=None):
        Group.__init__(self, name=self.top_group)
        self._larch = larch
        self._sys = None
        setattr(self, self.top_group, self)

        for gname in self.core_groups:
            thisgroup = Group(name=gname)
            if gname in GroupDocs:
                thisgroup.__doc__ = GroupDocs[gname]
            setattr(self, gname, thisgroup)

        self._sys.frames      = []
        self._sys.searchGroups = [self.top_group]
        self._sys.path        = ['.']
        self._sys.localGroup  = self
        self._sys.valid_commands = []
        self._sys.moduleGroup = self
        self._sys.__cache__  = [None]*4
        self._sys.saverestore_groups = []
        for grp in self.core_groups:
            self._sys.searchGroups.append(grp)
        self._sys.core_groups = tuple(self._sys.searchGroups[:])

        self._sys.modules = {'_main':self}
        for gname in self.core_groups:
            self._sys.modules[gname] = getattr(self, gname)
        self._fix_searchGroups()

        self._sys.config = Group(home_dir    = site_config.home_dir,
                                 history_file= site_config.history_file,
                                 init_files  = site_config.init_files,
                                 user_larchdir= site_config.user_larchdir,
                                 larch_version= site_config.larch_version,
                                 release_version = site_config.larch_release_version)

    def save_frame(self):
        " save current local/module group"
        self._sys.frames.append((self._sys.localGroup, self._sys.moduleGroup))

    def restore_frame(self):
        "restore last saved local/module group"
        try:
            lgrp, mgrp = self._sys.frames.pop()
            self._sys.localGroup = lgrp
            self._sys.moduleGroup  = mgrp
            self._fix_searchGroups()
        except:
            pass

    def set_frame(self, groups):
        "set current execution frame (localGroup, moduleGroup)"
        self._sys.localGroup, self._sys.moduleGroup  = groups
        self._fix_searchGroups()


    def _fix_searchGroups(self, force=False):
        """resolve list of groups to search for symbol names:

        The variable self._sys.searchGroups holds the list of group
        names for searching for symbol names.  A user can set this
        dynamically.  The names need to be absolute (that is, relative to
        _main, and can omit the _main prefix).

        This calclutes and returns self._sys.searchGroupObjects,
        which is the list of actual group objects (not names) resolved from
        the list of names in _sys.searchGroups)

        _sys.localGroup,_sys.moduleGroup come first in the search list,
        followed by any search path associated with that module (from
        imports for that module)
        """
        ##
        # check (and cache) whether searchGroups needs to be changed.
        sys = self._sys
        cache = sys.__cache__
        if len(cache) < 4:
            cache = [None]*4
        if (sys.localGroup   == cache[0] and
            sys.moduleGroup  == cache[1] and
            sys.searchGroups == cache[2] and
            cache[3] is not None and not force):
            return cache[3]

        if sys.moduleGroup is None:
            sys.moduleGroup = self.top_group
        if sys.localGroup is None:
            sys.localGroup = sys.moduleGroup

        cache[0] = sys.localGroup
        cache[1] = sys.moduleGroup
        snames  = []
        sgroups = []
        for grp in (sys.localGroup, sys.moduleGroup):
            if grp is not None and grp not in sgroups:
                sgroups.append(grp)
                snames.append(grp.__name__)

        sysmods = list(self._sys.modules.values())
        searchGroups  = sys.searchGroups[:]
        searchGroups.extend(self._sys.core_groups)
        for name in searchGroups:
            grp = None
            if name in self._sys.modules:
                grp = self._sys.modules[name]
            elif hasattr(self, name):
                gtest = getattr(self, name)
                if isinstance(gtest, Group):
                    grp = gtest
            elif '.' in name:
                parent, child= name.split('.')
                for sgrp in sysmods:
                    if (parent == sgrp.__name__ and
                        hasattr(sgrp, child)):
                        grp = getattr(sgrp, child)
                        break
            else:
                for sgrp in sysmods:
                    if hasattr(sgrp, name):
                        grp = getattr(sgrp, name)
                        break
            if grp is not None and grp not in sgroups:
                sgroups.append(grp)
                snames.append(name)

        self._sys.searchGroups = cache[2] = snames[:]
        sys.searchGroupObjects = cache[3] = sgroups[:]
        return sys.searchGroupObjects

    def get_parentpath(self, sym):
        """ get parent path for a symbol"""
        obj = self._lookup(sym)
        if obj is None:
            return
        out = []
        for s in reversed(self.__parents):
            if s.__name__ != '_main' or '_main' not in out:
                out.append(s.__name__)
        out.reverse()
        return '.'.join(out)

    def _lookup(self, name=None, create=False):
        """looks up symbol in search path
        returns symbol given symbol name,
        creating symbol if needed (and create=True)"""
        debug = False # not ('force'in name)
        if debug:
            print( '====\nLOOKUP ', name)
        searchGroups = self._fix_searchGroups()
        self.__parents = []
        if self not in searchGroups:
            searchGroups.append(self)

        def public_attr(grp, name):
            return (hasattr(grp, name)  and
                    not (grp is self and name in self._private))

        parts = name.split('.')
        if len(parts) == 1:
            for grp in searchGroups:
                if public_attr(grp, name):
                    self.__parents.append(grp)
                    return getattr(grp, name)

        # more complex case: not immediately found in Local or Module Group
        parts.reverse()
        top = parts.pop()
        out = self.__invalid_name
        if top == self.top_group:
            out = self
        else:
            for grp in searchGroups:
                if public_attr(grp, top):
                    self.__parents.append(grp)
                    out = getattr(grp, top)
        if out is self.__invalid_name:
            raise NameError(f"'{name}' is not defined")

        if len(parts) == 0:
            return out

        while parts:
            prt = parts.pop()
            if hasattr(out, prt):
                out = getattr(out, prt)
            elif create:
                val = None
                if len(parts) > 0:
                    val = Group(name=prt)
                setattr(out, prt, val)
                out = getattr(out, prt)
            else:
                raise LookupError(
                    f"cannot locate member '{prt}' of '{out}'")
        return out

    def has_symbol(self, symname):
        try:
            _ = self.get_symbol(symname)
            return True
        except (LookupError, NameError, ValueError):
            return False

    def has_group(self, gname):
        try:
            _ = self.get_group(gname)
            return True
        except (NameError, LookupError):
            return False

    def isgroup(self, sym):
        "test if symbol is a group"
        return isgroup(sym)

    def get_group(self, gname):
        "find group by name"
        sym = self._lookup(gname, create=False)
        if isgroup(sym):
            return sym
        raise LookupError(f"symbol '{gname}' found, but not a group")

    def create_group(self, **kw):
        "create a new Group, not placed anywhere in symbol table"
        return Group(**kw)

    def new_group(self, name, **kws):
        name = fixName(name)
        grp = Group(__name__ = name, **kws)
        self.set_symbol(name, value=grp)
        return grp

    def get_symbol(self, sym, create=False):
        "lookup and return a symbol by name"
        return self._lookup(sym, create=create)

    def set_symbol(self, name, value=None, group=None):
        "set a symbol in the table"
        grp = self._sys.localGroup
        if group is not None:
            grp = self.get_group(group)
        names = []

        for n in name.split('.'):
            if not isValidName(n):
                raise SyntaxError(f"invalid symbol name '{n}'")
            names.append(n)

        child = names.pop()
        for nam in names:
            if hasattr(grp, nam):
                grp = getattr(grp, nam)
                if not isgroup(grp):
                    raise ValueError(
                        f"cannot create subgroup of non-group '{grp}'")
            else:
                setattr(grp, nam, Group())

        setattr(grp, child, value)
        return value

    def del_symbol(self, name):
        "delete a symbol"
        sym = self._lookup(name, create=False)
        parent, child = self.get_parent(name)
        delattr(parent, child)

    def get_parent(self, name):
        """return parent group, child name for an absolute symbol name
        (as from _lookup) that is, a pair suitable for hasattr,
        getattr, or delattr
        """
        tnam = name.split('.')
        if len(tnam) < 1 or name == self.top_group:
            return (self, None)
        child = tnam.pop()
        sym = self
        if len(tnam) > 0:
            sym = self._lookup('.'.join(tnam))
        return sym, child

    def show_group(self, groupname):
        """display group members --- simple version for tests"""
        out = []
        try:
            group = self.get_group(groupname)
        except (NameError, LookupError):
            return 'Group %s not found' % groupname

        members = dir(group)
        out = ['f== {group.__name__}: {len(members)} symbols ==']
        for item in members:
            obj = getattr(group, item)
            dval = repr_value(obj)
            out.append(f'  {item}: {dval}')
        out.append('\n')
        self._larch.writer.write('\n'.join(out))



#!/usr/bin/env python
"""
general purpose file utilities
"""
from pathlib import Path
from random import Random

alphanum = 'abcdefghijklmnopqrstuvwxyz0123456789'

rng = Random()

def random_string(n, rng_seed=None):
    """  random_string(n)
    generates a random string of length n, that will match:
       [a-z][a-z0-9](n-1)
    """
    if rng_seed is not None:
        rng.seed(rng_seed)
    s = [nng.choice(alphanum[:26])]
    s.extend([rng.choice(alphanum) for i in range(n-2)])
    return ''.join(s)


def increment_filename(inpfile, ndigits=3, delim='.'):
    """
    increment a data filename, returning a new (non-existing) filename

       first see if a number is before '.'.  if so, increment it.
       second look for number in the prefix. if so, increment it.
       lastly, insert a '_001' before the '.', preserving suffix.

    the numerical part of the file name will contain at least three digits.

    >>> increment_filename('a.002')
    'a.003'
    >>> increment_filename('a.999')
    'a.1000'
    >>> increment_filename('b_017.xrf')
    'b_018.xrf'
    >>> increment_filename('x_10300243.dat')
    'x_10300244.dat'

    >>> increment_filename('x.dat')
    'x_001.dat'

    >>> increment_filename('C:/program files/oo/data/x.002')
    'C:/program files/ifeffit/data/x.003'

    >>> increment_filename('a_001.dat')
    'a_002.dat'

    >>> increment_filename('a.001.dat')
    'a.002.dat'

    >>> increment_filename('a_6.dat')
    'a_007.dat'

    >>> increment_filename('a_001.002')
    'a_001.003'

    >>> increment_filename("path/a.003")
    'path/a.004'
"""

    pinp = Path(inpfile)
    dirname, filename = pinp.parent.as_posix(), pinp.name
    base, ext = pinp.stem, pinp.suffix
    if ext == '':
        ext = '.000'

    if ext.startswith('.'):
        ext = ext[1:]
    ndigits = max(3, ndigits)

    def _incr(base, ext):
        if ext.isdigit():
            ext = f"{(int(ext)+1):0{ndigits}d}"
        else:
            found = False
            if '_' in base:
                parts = base.split('_')
                for iw, word in enumerate(parts[::-1]):
                    if word.isdigit():
                        parts[len(parts)-iw-1] = f"{(int(word)+1):0{ndigits}d}"
                        found = True
                        break
                base = '_'.join(parts)
            if not found and '.' in base:
                parts = base.split('.')
                for iw, word in enumerate(parts[::-1]):
                    if word.isdigit():
                        parts[len(parts)-iw-1] = f"{(int(word)+1):0{ndigits}d}"
                        found = True
                        break
                base = '.'.join(parts)
            if not found:
                base = f"{base}_001"
        return (base, ext)

    # increment once
    base, ext = _incr(base, ext)
    fout = Path(dirname,f"{base}{delim}{ext}")

    # then gaurantee that file does not exist,
    # continuing to increment if necessary
    while fout.exists():
        base, ext = _incr(base, ext)
        fout = Path(dirname,f"{base}{delim}{ext}")
    return fout.as_posix()

def new_filename(fname=None, ndigits=3):
    """ generate a new file name, either based on
    filename or generating a random one

    >>> new_filename(fname='x.001')
    'x.002'
    # if 'x.001' exists
    """
    if fname is None:
        fname = f"{random_string(6)}.{1:0{ndigits}d}"

    if Path(fname).exists():
        fname = increment_filename(fname, ndigits=ndigits)
    return fname

def new_dirname(dirname=None, ndigits=3):
    """ generate a new subdirectory name (no '.' in name), either
    based on dirname or generating a random one

    >>> new_dirname('x.001')
    'x_002'
    # if 'x_001' exists
    """
    if dirname is None:
        dirname = f"{random_string(6)}_{1:0{ndigits}d}"

    dirname = dirname.replace('.', '_')
    if Path(dirname).exists():
        dirname = increment_filename(dirname, ndigits=ndigits, delim='_')
    return dirname

def test_incrementfilename():
    tests = (('a.002', 'a.003'),
             ('a.999', 'a.1000'),
             ('b_017.xrf',  'b_018.xrf'),
             ('x_10300243.dat', 'x_10300244.dat'),
             ('x.dat' , 'x_001.dat'),
             ('C:/program files/data/x.002',
              'C:/program files/data/x.003'),
             ('a_001.dat', 'a_002.dat'),
             ('a_6.dat', 'a_007.dat'),
             ('a_001.002', 'a_001.003'),
             ('path/a.003',  'path/a.004'))
    npass = nfail = 0
    for inp,out in tests:
        tval = increment_filename(inp)
        if tval != out:
            print(f"Error converting {inp}")
            print(f"Got '{tval}'  expected '{out}'")
            nfail = nfail + 1
        else:
            npass = npass + 1
    print(f'Passed {npass} of {npass+nfail} tests')



from pathlib import Path
from larch.io import read_ascii, guess_beamline, guess_filereader, read_fdmnes

base_dir = Path(__file__).parent.parent.resolve()

def _tester(fname, return_group=False):
    fname =  base_dir / 'examples' / 'xafsdata' / 'beamlines' / fname
    group = read_ascii(fname)
    cls = guess_beamline(group.header)
    bldat = cls(group.header)
    labels = bldat.get_array_labels()
    print(fname, cls.__name__, len(labels), group.data.shape, labels)
    if return_group:
        return bldat, labels, group
    else:
        return bldat, labels

def test_apsxsd_new(fname='APS9BM_2019.dat'):
    bldat, labels = _tester(fname)
    assert('aps xsd' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(labels[0]  == 'mono_energy')
    assert(labels[1].startswith('scaler_pre'))
    assert(labels[4].startswith('i0'))


def test_apsxsd20id_newer(fname='APS20ID_2022.dat'):
    bldat, labels = _tester(fname)
    assert('aps xsd' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(labels[0].startswith('mono_energy'))
    assert(labels[1].startswith('scaler_pre'))
    assert(labels[-1] == "xsp3_4_total")
    assert('it' in labels)


def test_apsxsd20id_new(fname='APS20ID_2018.dat'):
    bldat, labels = _tester(fname)
    assert('aps xsd' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(labels[0].startswith('mono_energy'))
    assert(labels[1].startswith('scaler_pre'))
    assert('it' in labels)


def test_apsxsd_old(fname='APS20BM_2001.dat'):
    bldat, labels = _tester(fname)
    assert('aps xsd' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(labels[0]  == 'mono_energy')
    assert(labels[1].startswith('scaler_pre'))
    assert(labels[2] == 'i0')


def test_apsmrcat(fname='APS10BM_2019.dat'):
    bldat, labels = _tester(fname)
    assert('aps mrcat' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) == 5)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'io')
    assert(labels[2]  == 'it')
    assert(labels[3]  == 'iref')

def test_apsgse(fname='APS13ID_2019.dat'):
    bldat, labels = _tester(fname)
    assert('gse epicsscan' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) == 21)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'tscaler')
    assert(labels[2]  == 'i0')
    assert(labels[3]  == 'i1')
    assert(labels[4]  == 'i2')
    assert(labels[5]  == 'mn_ka_mca1')

def test_apsgse_old(fname='APS13ID_2008.dat'):
    bldat, labels = _tester(fname)
    assert('gse epicsscan' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) == 4)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'scaler_count_time')
    assert(labels[2]  == 'i0')
    assert(labels[3]  == 'i1')

def test_aps12bm(fname='APS12BM_2019.dat'):
    bldat, labels = _tester(fname)
    assert('12bm' in bldat.name.lower())
    assert(1 == bldat.energy_column)
    assert('keV' == bldat.energy_units)
    assert(len(labels) == 26)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'sec')
    assert(labels[2]  == 'mononrg')
    assert(labels[3]  == 'i0')
    assert(labels[4]  == 'i1')
    assert(labels[5]  == 'i2')

def test_aps9bm2006(fname='APS9BM_2006.dat'):
    bldat, labels = _tester(fname)
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 5)
    assert(labels[0]  == 'energy')
    assert('i0' in labels)

def test_esrfsnbl(fname='ESRF_SNBL_2013.dat'):
    bldat, labels = _tester(fname)
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 5)
    assert(labels[0]  == 'mon')
    assert(labels[1]  == 'det1')
    assert(labels[2]  == 'det2')

def test_nsls2_6bm(fname='NSLS6BM_2019.dat'):
    bldat, labels = _tester(fname)
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 5)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'requested_energy')
    assert(labels[2]  == 'measurement_time')
    assert(labels[3]  == 'xmu')
    assert(labels[4]  == 'i0')

def test_nsls2_8id(fname='NSLS8ID_2019.dat'):
    bldat, labels = _tester(fname)
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 5)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'i0')
    assert(labels[2]  == 'ir')
    assert(labels[3]  == 'it')
    assert(labels[4]  == 'iff')

def test_ssrl1(fname='SSRL1_2006.dat'):
    bldat, labels = _tester(fname)
    assert(3 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) == 6)
    assert(labels[0]  == 'real_time_clock')
    assert(labels[1]  == 'requested_energy')
    assert(labels[2]  == 'achieved_energy')
    assert(labels[3]  == 'i0')
    assert(labels[4]  == 'i1')
    assert(labels[5]  == 'i2')

def test_ssrl2(fname='SSRLmicro_2008.dat'):
    bldat, labels = _tester(fname)
    assert(2 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 65)
    assert(labels[0]  == 'real_time_clock')
    assert(labels[1]  == 'requested_energy')
    assert(labels[2]  == 'i0')
    assert(labels[3]  == 'i1')
    assert(labels[4]  == 'i2')
    assert(labels[5]  == 'sca1_1')

def test_nslsxdac(fname='NSLS_XDAC_2011.dat'):
    bldat, labels = _tester(fname)
    assert(1 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 15)
    assert(labels[0]  == 'energy')
    assert(labels[1]  == 'i0')
    assert(labels[2]  == 'it')
    assert(labels[3]  == 'ifch1')
    assert(labels[4]  == 'ifch2')

def test_clshxma(fname='CLSHXMA.dat'):
    bldat, labels = _tester(fname)
    assert(4 == bldat.energy_column)
    assert('eV' == bldat.energy_units)
    assert(len(labels) > 10)
    assert(labels[0]  == 'event_id')
    assert(labels[1]  == 'absenergyname')
    assert(labels[2]  == 'energyfeedback')
    assert(labels[3]  == 'energyachieved')
    assert(labels[4]  == 'detector1')

def test_kekpf12c(fname='PFBL12C_2005.dat'):
    bldat, labels = _tester(fname)
    assert(2 == bldat.energy_column)
    assert('deg' == bldat.energy_units)
    assert(bldat.mono_dspace > 3)
    assert(len(labels) == 5)

def test_one_line_header(fname='ESRF_BM08_LISA_2021.dat'):
    bldat, labels = _tester(fname)
    assert(labels == ['ebraggenergy', 'i0_eh1', 'i1_eh1', 'mu', 'i1_eh2', 'ir_eh2', 'mu_ref'])

def test_zero_line_header(fname='generic_columns_no_header.dat'):
    bldat, labels, group = _tester(fname, return_group=True)
    assert(group.array_labels == ['col1', 'col2', 'col3', 'col4', 'col5', 'col6', 'col7'])

def test_fdmnes(fnames=['FDMNES_2022_Mo2C_out.dat', 'FDMNES_2022_Mo2C_out_conv.dat']):
    for fname in fnames:
        fname =  base_dir / 'examples' / 'xafsdata' / 'beamlines' / fname
        assert(guess_filereader(fname) == 'read_fdmnes')
        group = read_fdmnes(fname)
        assert(group.array_labels == ['energy', 'xanes'])
        assert(group.header_dict['E_edge'] == 20000.0)

if __name__ == '__main__':
    test_apsxsd_new()
    test_apsxsd_old()
    test_apsgse()
    test_apsgse_old()
    test_apsmrcat()
    test_aps12bm()
    test_aps9bm2006()
    test_esrfsnbl()
    test_nsls2_8id()
    test_nsls2_6bm()
    test_ssrl1()
    test_ssrl2()
    test_nslsxdac()
    test_clshxma()
    test_kekpf12c()
    test_one_line_header()
    test_zero_line_header()
    test_fdmnes()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
RIXS data object
================
"""
import numpy as np
import copy
import time
from itertools import cycle
from scipy.interpolate import griddata
from silx.io.dictdump import dicttoh5, h5todict

from larch.math.gridxyz import gridxyz
from larch.xafs.xafsutils import guess_energy_units
from larch.utils.logging import getLogger

_logger = getLogger(__name__)  #: module logger
_logger.setLevel("INFO")


def _tostr(arr):
    """Numpy array to string"""
    try:
        return np.array_str(arr)
    except Exception:
        return arr


def _restore_from_array(dictin):
    """restore str/float from a nested dictionary of numpy.ndarray (when using silx.io.dictdump.h5todict)

    Note: discussed here https://github.com/silx-kit/silx/issues/3633
    """
    for k, v in dictin.items():
        if isinstance(v, dict):
            _restore_from_array(v)
        else:
            if isinstance(v[()], np.str_):
                dictin[k] = np.array_str(v)
            if isinstance(v[()], (np.float64, np.float32)):
                dictin[k] = copy.deepcopy(v.item())


class CycleColors:
    """Utility for setting the line colors of the RIXS map cuts"""

    DEFAULT_PALETTE = (
        "#1F77B4",
        "#AEC7E8",
        "#FF7F0E",
        "#FFBB78",
        "#2CA02C",
        "#98DF8A",
        "#D62728",
        "#FF9896",
        "#9467BD",
        "#C5B0D5",
        "#8C564B",
        "#C49C94",
        "#E377C2",
        "#F7B6D2",
        "#7F7F7F",
        "#C7C7C7",
        "#BCBD22",
        "#DBDB8D",
        "#17BECF",
        "#9EDAE5",
    )

    def __init__(self) -> None:
        self.colors = cycle(self.DEFAULT_PALETTE)

    def get_color(self) -> None:
        return next(self.colors)


class RixsData(object):
    """RIXS plane object"""

    def __init__(self, name=None, logger=None):
        """Constructor"""

        self.__name__ = "RixsData_{0}".format(hex(id(self)))
        self.name = name or self.__name__
        self.label = self.name

        self._logger = logger or _logger
        self._palette = CycleColors()
        self._no_save = ("_logger", "_palette")

        self.sample_name = "UnknownSample"
        self.counter_all, self.counter_signal, self.counter_norm = None, None, None
        self._x, self._y, self._z = None, None, None
        self.ene_in, self.ene_out, self.rixs_map = None, None, None
        self.ene_et, self.rixs_et_map = None, None
        self.ene_grid, self.ene_unit, self.grid_method = None, None, None
        self.line_cuts = {}
        self.datatype = "rixs"

    def set_energy_unit(self, unit=None):
        """set the energy unit to eV"""
        if unit is not None:
            self.ene_unit = unit
        if self.ene_unit is None:
            self.ene_unit = guess_energy_units(self._x)
        if self.ene_unit == "keV":
            self._logger.info(f"Energy unit is {self.ene_unit} -> converting to eV")
            self._x *= 1000
            self._y *= 1000
            self.ene_grid = 0.1
            self.reset()
            self.ene_unit = "eV"
        assert (
            self.ene_unit == "eV"
        ), f"energy unit is {self.set_energy_unit} -> must be eV"

    def load_from_dict(self, rxdict):
        """Load RIXS data from a dictionary

        Parameters
        ----------
        rxdict : dict
            Minimal required structure
            {
             'writer_version': '1.5.0',
             'sample_name': str,
             '_x': 1D array,  #: energy in
             '_y': 1D array,  #: energy out
             '_z': 1D array,  #: signal
            }

        Return
        ------
        None, set attributes: self.*
        """
        self.__dict__.update(rxdict)
        self.set_energy_unit()
        self.grid_rixs_from_col()

    def load_from_h5(self, filename):
        """Load RIXS from HDF5 file"""
        rxdict = h5todict(filename)
        _restore_from_array(rxdict)
        if not ("writer_version" in rxdict.keys()):
            self._logger.error("Key 'writer_version' not found")
            return
        if not ("1.5" in _tostr(rxdict["writer_version"])):
            self._logger.warning("Data format not understood")
            return
        self.load_from_dict(rxdict)
        self._logger.info("RIXS map loaded from file: {0}".format(filename))

    def load_from_ascii(self, filename, **kws):
        """load data from a 3 columns ASCII file assuming the format:

        e_in(eV), e_out(eV), signal

        """

        try:
            dat = np.loadtxt(filename)
            self.filename = filename
            self._logger.info("Loaded {0}".format(filename))
        except Exception:
            self._logger.error("Cannot load from {0}".format(filename))
            return

        self._x = dat[:, 0]
        self._y = dat[:, 1]
        self._z = dat[:, 2]

        self.set_energy_unit()
        self.reset()

    def save_to_h5(self, filename=None):
        """Dump dictionary representation to HDF5 file"""
        if filename is None:
            filename = f"{self.filename.split('.')[0]}.h5"
        save_dict = copy.deepcopy(self.__dict__)
        for dkey in self._no_save:
            try:
                del save_dict[dkey]
            except KeyError:
                continue
        dicttoh5(save_dict, filename, update_mode="replace")
        self._logger.info(f"{self.name} saved to {filename}")

    def crop(self, crop_area, et=None):
        """Crop the plane in a given range

        Parameters
        ----------

        crop_area : tuple
            (x1, y1, x2, y2) : floats
            x1 < x2 (ene_in)
            y1 < y2 (if yet=False: ene_out, else: ene_et)

        et: bool,
            if True: y1, y2 are given in energy transfer

        """
        self._crop_area = crop_area
        x1, y1, x2, y2 = crop_area
        assert x1 < x2, "wrong crop area, x1 >= x2"
        assert y1 < y2, "wrong crop area, y1 >= y2"

        if et is None:
            if y2 < np.max(self.ene_et):
                self._logger.debug("crop in energy transfer")
                et = True
            else:
                self._logger.debug("crop in emission energy")
                et = False

        _xystep = self.ene_grid or 0.1
        _method = self.grid_method or "linear"

        _nxpts = int((x2 - x1) / _xystep)
        _xcrop = np.linspace(x1, x2, num=_nxpts)

        if et:
            _etmin = y1
            _etmax = y2
            _ymin = x1 - _etmax
            _ymax = x2 - _etmin
            self._logger.debug(f"-> emission range: {_ymin:.2f}:{_ymax:.2f}")
        else:
            _ymin = y1
            _ymax = y2
            _etmin = x2 - _ymax
            _etmax = x1 - _ymin
            self._logger.debug(f"-> et range: {_etmin:.2f}:{_etmax:.2f}")

        _netpts = int((_etmax - _etmin) / _xystep)
        _nypts = int((_ymax - _ymin) / _xystep)
        _etcrop = np.linspace(_etmin, _etmax, num=_netpts)
        _ycrop = np.linspace(_ymin, _ymax, num=_nypts)


        _xx, _yy = np.meshgrid(_xcrop, _ycrop)
        _exx, _et = np.meshgrid(_xcrop, _etcrop)
        self._logger.info("Gridding data...")
        _zzcrop = griddata((self._x, self._y), self._z, (_xx, _yy), method=_method)
        _ezzcrop = griddata(
            (self._x, self._x - self._y), self._z, (_exx, _et), method=_method
        )

        self.ene_in = _xcrop
        self.ene_out = _ycrop
        self.ene_et = _etcrop
        self.rixs_map = _zzcrop
        self.rixs_et_map = _ezzcrop
        self.label = f"{self.name} [{self._crop_area}]"

    def reset(self, **grid_kws):
        """resets to initial data"""
        self._logger.info("resetting to initial data (grid RIXS plane and line cuts)")
        self.grid_rixs_from_col(**grid_kws)
        self.line_cuts = {}
        self.label = self.name
        self._palette = None
        self._palette = CycleColors()

    def grid_rixs_from_col(self, ene_grid=None, grid_method=None):
        """Grid RIXS map from XYZ columns"""
        if ene_grid is not None:
            self.ene_grid = ene_grid
        if grid_method is not None:
            self.grid_method = grid_method
        _xystep = self.ene_grid or 0.1
        _method = self.grid_method or "linear"
        self.ene_in, self.ene_out, self.rixs_map = gridxyz(
            self._x, self._y, self._z, xystep=_xystep, method=_method
        )
        self._et = self._x - self._y
        _, self.ene_et, self.rixs_et_map = gridxyz(
            self._x, self._et, self._z, xystep=_xystep, method=_method
        )
        self.ene_grid = _xystep

    def cut(self, energy=None, mode="CEE", label=None):
        """cut the RIXS plane at a given energy

        Parameters
        ----------
        energy : float
            energy of the cut

        mode : str
            defines the way to cut the plane:
                - "CEE" (constant emission energy)
                - "CIE" (constant incident energy)
                - "CET" (constant energy transfer)

        label : str, optional [None]
            custom label, if None: label = 'mode_enecut'

        Return
        ------
            None -> adds dict(x:array, y:array, info:dict) to self.lcuts[cut_key]:dict, where

            info = {label: str,     #: 'mode_enecut'
                    mode: str,      #: as input
                    enecut: float,  #: energy cut given from the initial interpolation
                    datatype: str,  #: 'xas' or 'xes'
                    color: str,     #: color from a common palette
                    timestamp: str, #: time stamp
                    }
        """
        assert energy is not None, "The energy of the cut must be given"

        mode = mode.upper()

        if mode == "CEE":
            xc = self.ene_in
            iy = np.abs(self.ene_out - energy).argmin()
            enecut = self.ene_out[iy]
            yc = self.rixs_map[iy, :]
            datatype = "xas"
        elif mode == "CIE":
            xc = self.ene_out
            iy = np.abs(self.ene_in - energy).argmin()
            enecut = self.ene_in[iy]
            yc = self.rixs_map[:, iy]
            datatype = "xes"
        elif mode == "CET":
            xc = self.ene_in
            iy = np.abs(self.ene_et - energy).argmin()
            enecut = self.ene_et[iy]
            yc = self.rixs_et_map[iy, :]
            datatype = "xas"
        else:
            self._logger.error(f"wrong mode: {mode}")
            return

        if label is None:
            label = f"{mode}_{enecut:.1f}"

        info = dict(
            label=label,
            mode=mode,
            enecut=enecut,
            datatype=datatype,
            color=self._palette.get_color(),
            timestamp="{0:04d}-{1:02d}-{2:02d}_{3:02d}{4:02d}".format(
                *time.localtime()
            ),
        )

        cut_key = f"{mode}_{enecut*10:.0f}"
        self.line_cuts[cut_key] = dict(x=xc, y=yc, info=info)
        self._logger.info(f"added RIXS {mode} cut: '{label}'")

    def norm(self):
        """Simple map normalization to max-min"""
        self.rixs_map = self.rixs_map / (
            np.nanmax(self.rixs_map) - np.nanmin(self.rixs_map)
        )
        self._logger.info("rixs map normalized to max-min")


if __name__ == "__main__":
    pass




import os

from six.moves.xmlrpc_client import ServerProxy
lserve = ServerProxy('http://localhost:4966')

# print lserve.system.listMethods()

parent, thisdir = os.path.split(os.path.abspath(os.curdir))

lserve.chdir(os.path.join(parent, 'feffit'))

lserve.larch("run('doc_feffit1.lar')")

print('## Messages: ')
print(lserve.get_messages())



import wx
import sys
from collections import namedtuple
from pyshortcuts import platform
from ..version import check_larchversion
from . import GridPanel, SimpleText, OkCancel, LEFT, HLine, Button


class LarchUpdaterDialog(wx.Dialog):
    """dialog for running larch updates"""
    def __init__(self, parent, caller='this program', **kws):
        title = "Checking for Larch updates..."
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title, size=(525, 250))

        vinfo = check_larchversion()
        self.update_available = vinfo.update_available

        upgrade_pycmd = f'{sys.executable} -m pip install --upgrade xraylarch'
        panel = GridPanel(self, ncols=3, nrows=4, pad=4, itemstyle=LEFT)
        panel.Add((5, 5))

        def addline(text):
            panel.Add(SimpleText(panel, text), newrow=True)

        def toclipboard(event=None, **kws):
            cb = wx.TheClipboard
            if cb.IsOpened():
                cb.Close()
            cb.Open()
            cb.SetData(wx.TextDataObject(upgrade_pycmd))
            cb.Close()

        addline(' Installed Version: {:s}'.format(vinfo.local_version))
        addline(' Latest Version: {:s}'.format(vinfo.remote_version))
        panel.Add((5, 5))

        if not vinfo.update_available:
            addline(' Your version is up to date!')
        else:
            if platform.startswith('win'):
                addline(f' Close {caller} and Run "Larch Updater" from your Larch Desktop Folder')
            else:
                addline(f' Click OK to update (you will need to restart {caller})')
            panel.Add((5, 5))
            addline(' You can also update from a Terminal with:')
            addline(f' {upgrade_pycmd}')
            panel.Add(Button(panel, 'Copy Upgrade Command To Clipboard',
                             action=toclipboard), newrow=True)
            panel.Add((5, 5), newrow=True)

        panel.Add(HLine(panel, size=(240, 3)), dcol=2, newrow=True)
        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()

    def GetResponse(self):
        self.Raise()
        out = namedtuple('UpgradeResponse', ('ok', 'run_updates'))
        result = out((wx.ID_OK==self.ShowModal()), self.update_available)
        if platform.startswith('win'):
            result = out(False, self.update_available)
        return result



#!/usr/bin/env python
'''
build American Mineralogist Crystal Structure Databse (amcsd)
'''

import os
import requests
import numpy as np
from itertools import groupby
import larch
from .xrd_fitting import peaklocater
from .xrd_cif import create_xrdcif, SPACEGROUPS
from .xrd_tools import lambda_from_E

import json
from larch.utils.jsonutils import encode4js, decode4js

from sqlalchemy import (create_engine, MetaData, Table, Column, Integer,
                        String, Unicode, PrimaryKeyConstraint,
                        ForeignKeyConstraint, ForeignKey, Numeric, func,
                        and_, or_, not_, tuple_)

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.pool import SingletonThreadPool

SYMMETRIES = ['triclinic', 'monoclinic', 'orthorhombic', 'tetragonal',
              'trigonal', 'hexagonal', 'cubic']

ELEMENTS = [['1',  'Hydrogen',  'H'], ['2',  'Helium',  'He'], ['3',  'Lithium',  'Li'],
            ['4',  'Beryllium',  'Be'], ['5',  'Boron',  'B'], ['6',  'Carbon',  'C'],
            ['7',  'Nitrogen',  'N'], ['8',  'Oxygen',  'O'], ['9',  'Fluorine',  'F'],
            ['10',  'Neon',  'Ne'], ['11',  'Sodium',  'Na'], ['12',  'Magnesium',  'Mg'],
            ['13',  'Aluminum',  'Al'], ['14',  'Silicon',  'Si'], ['15',  'Phosphorus',  'P'],
            ['16',  'Sulfur',  'S'], ['17',  'Chlorine',  'Cl'], ['18',  'Argon',  'Ar'],
            ['19',  'Potassium',  'K'], ['20',  'Calcium',  'Ca'], ['21',  'Scandium',  'Sc'],
            ['22',  'Titanium',  'Ti'], ['23',  'Vanadium',  'V'], ['24',  'Chromium',  'Cr'],
            ['25',  'Manganese',  'Mn'], ['26',  'Iron',  'Fe'], ['27',  'Cobalt',  'Co'],
            ['28',  'Nickel',  'Ni'], ['29',  'Copper',  'Cu'], ['30',  'Zinc',  'Zn'],
            ['31',  'Gallium',  'Ga'], ['32',  'Germanium',  'Ge'], ['33',  'Arsenic',  'As'],
            ['34',  'Selenium',  'Se'], ['35',  'Bromine',  'Br'], ['36',  'Krypton',  'Kr'],
            ['37',  'Rubidium',  'Rb'], ['38',  'Strontium',  'Sr'], ['39',  'Yttrium',  'Y'],
            ['40',  'Zirconium',  'Zr'], ['41',  'Niobium',  'Nb'], ['42',  'Molybdenum',  'Mo'],
            ['43',  'Technetium',  'Tc'], ['44',  'Ruthenium',  'Ru'], ['45',  'Rhodium',  'Rh'],
            ['46',  'Palladium',  'Pd'], ['47',  'Silver',  'Ag'], ['48',  'Cadmium',  'Cd'],
            ['49',  'Indium',  'In'], ['50',  'Tin',  'Sn'], ['51',  'Antimony',  'Sb'],
            ['52',  'Tellurium',  'Te'], ['53',  'Iodine',  'I'], ['54',  'Xenon',  'Xe'],
            ['55',  'Cesium',  'Cs'], ['56',  'Barium',  'Ba'], ['57',  'Lanthanum',  'La'],
            ['58',  'Cerium',  'Ce'], ['59',  'Praseodymium',  'Pr'], ['60',  'Neodymium',  'Nd'],
            ['61',  'Promethium',  'Pm'], ['62',  'Samarium',  'Sm'], ['63',  'Europium',  'Eu'],
            ['64',  'Gadolinium',  'Gd'], ['65',  'Terbium',  'Tb'], ['66',  'Dysprosium',  'Dy'],
            ['67',  'Holmium',  'Ho'], ['68',  'Erbium',  'Er'], ['69',  'Thulium',  'Tm'],
            ['70',  'Ytterbium',  'Yb'], ['71',  'Lutetium',  'Lu'], ['72',  'Hafnium',  'Hf'],
            ['73',  'Tantalum',  'Ta'], ['74',  'Tungsten',  'W'], ['75',  'Rhenium',  'Re'],
            ['76',  'Osmium',  'Os'], ['77',  'Iridium',  'Ir'], ['78',  'Platinum',  'Pt'],
            ['79',  'Gold',  'Au'], ['80',  'Mercury',  'Hg'], ['81',  'Thallium',  'Tl'],
            ['82',  'Lead',  'Pb'], ['83',  'Bismuth',  'Bi'], ['84',  'Polonium',  'Po'],
            ['85',  'Astatine',  'At'], ['86',  'Radon',  'Rn'], ['87',  'Francium',  'Fr'],
            ['88',  'Radium',  'Ra'], ['89',  'Actinium',  'Ac'], ['90',  'Thorium',  'Th'],
            ['91',  'Protactinium',  'Pa'], ['92',  'Uranium',  'U'], ['93',  'Neptunium',  'Np'],
            ['94',  'Plutonium',  'Pu'], ['95',  'Americium',  'Am'], ['96',  'Curium',  'Cm'],
            ['97',  'Berkelium',  'Bk'], ['98',  'Californium',  'Cf'], ['99',  'Einsteinium',  'Es'],
            ['100',  'Fermium',  'Fm'], ['101',  'Mendelevium',  'Md'], ['102',  'Nobelium',  'No'],
            ['103',  'Lawrencium',  'Lr'], ['104',  'Rutherfordium',  'Rf'], ['105',  'Dubnium',  'Db'],
            ['106',  'Seaborgium',  'Sg'], ['107',  'Bohrium',  'Bh'], ['108',  'Hassium',  'Hs'],
            ['109',  'Meitnerium',  'Mt'], ['110',  'Darmstadtium',  'Ds'], ['111',  'Roentgenium',  'Rg'],
            ['112',  'Ununbium',  'Uub'], ['113',  'Ununtrium',  'Uut'], ['114',  'Ununquadium',  'Uuq'],
            ['115',  'Ununpentium',  'Uup'], ['116',  'Ununhexium',  'Uuh'], ['117',  'Ununseptium',  'Uus'],
            ['118',  'Ununoctium',  'Uuo']]

CATEGORIES = ['soil',
              'salt',
              'clay']

QMIN = 0.2
QMAX = 10.0
QSTEP = 0.01
QAXIS = np.arange(QMIN, QMAX+QSTEP, QSTEP)

ENERGY = 19000 ## units eV
_cifdb = None

def get_cifdb(dbname='amcsd_cif0.db', _larch=None):
    global _cifdb
    if _cifdb is None:
        _cifdb = cifDB(dbname=dbname)
    if _larch is not None:
        symname = '_xray._cifdb'
        if not _larch.symtable.has_symbol(symname):
            _larch.symtable.set_symbol(symname, _cifdb)
    return _cifdb

def make_engine(dbname):
    return create_engine('sqlite:///%s' % (dbname),
                         poolclass=SingletonThreadPool)

def iscifDB(dbname):
    '''
    test if a file is a valid scan database:
    must be a sqlite db file, with tables named according to _tables
    '''
    _tables = ('ciftbl',
               'elemtbl',
               'nametbl',
               #'formtbl',
               'spgptbl',
               'symtbl',
               'authtbl',
               'qtbl',
               'cattbl',
               'symref',
               #'compref',
               #'qref',
               'authref',
               'catref')
    result = False
    try:
        engine = make_engine(dbname)
        meta = MetaData()
        meta.reflect(bind=engine)
        result = all([t in meta.tables for t in _tables])
    except:
        pass
    return result


class cifDB(object):
    '''
    interface to the American Mineralogist Crystal Structure Database
    '''
    def __init__(self, dbname=None, read_only=True,verbose=False):

        ## This needs to be modified for creating new if does not exist.
        self.version = '0.0.2'
        self.dbname = dbname
        if verbose:
            print('\n\n================ %s ================\n' % self.dbname)
        if not os.path.exists(self.dbname):
            parent, child = os.path.split(__file__)
            self.dbname = os.path.join(parent, self.dbname)
            if not os.path.exists(self.dbname):
                print("File '%s' not found; building a new database!" % self.dbname)
                self.create_cifdb(name=self.dbname)
            else:
                if not iscifDB(self.dbname):
                    raise ValueError("'%s' is not a valid cif database file!" % self.dbname)

        self.dbname = self.dbname
        self.engine = make_engine(self.dbname)
        self.conn = self.engine.connect()

        kwargs = {}
        if read_only:
            kwargs = {'autoflush': True, 'autocommit':False}
            def readonly_flush(*args, **kwargs):
                return
            self.session = sessionmaker(bind=self.engine, **kwargs)()
            self.session.flush = readonly_flush
        else:
            self.session = sessionmaker(bind=self.engine, **kwargs)()

        self.metadata =  MetaData()
        self.metadata.reflect(bind=self.engine)
        tables = self.tables = self.metadata.tables

        ## Load tables
        self.elemtbl = Table('elemtbl', self.metadata)
        self.nametbl = Table('nametbl', self.metadata)
        self.formtbl = Table('formtbl', self.metadata)
        self.spgptbl = Table('spgptbl', self.metadata)
        self.symtbl  = Table('symtbl',  self.metadata)
        self.authtbl = Table('authtbl', self.metadata)
        self.qtbl    = Table('qtbl',    self.metadata)
        self.cattbl  = Table('cattbl',  self.metadata)

        self.symref  = Table('symref',  self.metadata)
        self.compref = Table('compref', self.metadata)
        self.qref    = Table('qref',    self.metadata)
        self.authref = Table('authref', self.metadata)
        self.catref  = Table('catref',  self.metadata)

        self.ciftbl  = Table('ciftbl', self.metadata)

        self.axis = np.array([float(q[0]) for q in self.query(self.qtbl.c.q).all()])


    def query(self, *args, **kws):
        "generic query"
        return self.session.query(*args, **kws)

    def close(self):
        "close session"
        self.session.flush()
        self.session.close()

    def create_cifdb(self,name=None,verbose=False):

        if name is None:
            self.dbname = 'amcsd_cif0.db'
            counter = 0
            while os.path.exists(self.dbname % counter):
                counter += 1
            self.dbname = self.dbname % counter
        else:
            self.dbname = name

        self.open_database()

        ###################################################
        ## Look up tables
        elemtbl = Table('elemtbl', self.metadata,
                  Column('z', Integer, primary_key=True),
                  Column('element_name', String(40), unique=True, nullable=True),
                  Column('element_symbol', String(2), unique=True, nullable=False)
                  )
        nametbl = Table('nametbl', self.metadata,
                  Column('mineral_id', Integer, primary_key=True),
                  Column('mineral_name', String(30), unique=True, nullable=True)
                  )
        formtbl = Table('formtbl', self.metadata,
                  Column('formula_id', Integer, primary_key=True),
                  Column('formula_name', String(30), unique=True, nullable=True)
                  )
        spgptbl = Table('spgptbl', self.metadata,
                  Column('iuc_id', Integer),
                  Column('hm_notation', String(16), unique=True, nullable=True),
                  PrimaryKeyConstraint('iuc_id', 'hm_notation')
                  )
        symtbl  = Table('symtbl', self.metadata,
                  Column('symmetry_id', Integer, primary_key=True),
                  Column('symmetry_name', String(16), unique=True, nullable=True)
                  )
        authtbl = Table('authtbl', self.metadata,
                  Column('author_id', Integer, primary_key=True),
                  Column('author_name', String(40), unique=True, nullable=True)
                  )
        qtbl    = Table('qtbl', self.metadata,
                  Column('q_id', Integer, primary_key=True),
                  #Column('q', Float()) ## how to make this work? mkak 2017.02.14
                  Column('q', String())
                  )
        cattbl  = Table('cattbl', self.metadata,
                  Column('category_id', Integer, primary_key=True),
                  Column('category_name', String(16), unique=True, nullable=True)
                  )
        ###################################################
        ## Cross-reference tables
        symref  = Table('symref', self.metadata,
                  Column('iuc_id', None, ForeignKey('spgptbl.iuc_id')),
                  Column('symmetry_id', None, ForeignKey('symtbl.symmetry_id')),
                  PrimaryKeyConstraint('iuc_id', 'symmetry_id')
                  )
        compref = Table('compref', self.metadata,
                  Column('z', None, ForeignKey('elemtbl.z')),
                  Column('amcsd_id', None, ForeignKey('ciftbl.amcsd_id')),
                  PrimaryKeyConstraint('z', 'amcsd_id')
                  )
        qref    = Table('qref', self.metadata,
                  Column('q_id', None, ForeignKey('qtbl.q_id')),
                  Column('amcsd_id', None, ForeignKey('ciftbl.amcsd_id')),
                  PrimaryKeyConstraint('q_id', 'amcsd_id')
                  )
        authref = Table('authref', self.metadata,
                  Column('author_id', None, ForeignKey('authtbl.author_id')),
                  Column('amcsd_id', None, ForeignKey('ciftbl.amcsd_id')),
                  PrimaryKeyConstraint('author_id', 'amcsd_id')
                  )
        catref  = Table('catref', self.metadata,
                  Column('category_id', None, ForeignKey('cattbl.category_id')),
                  Column('amcsd_id', None, ForeignKey('ciftbl.amcsd_id')),
                  PrimaryKeyConstraint('category_id', 'amcsd_id')
                  )
        ###################################################
        ## Main table
        ciftbl  = Table('ciftbl', self.metadata,
                  Column('amcsd_id', Integer, primary_key=True),
                  Column('mineral_id', Integer),
                  Column('formula_id', Integer),
                  Column('iuc_id', ForeignKey('spgptbl.iuc_id')),
                  Column('a', String(5)),
                  Column('b', String(5)),
                  Column('c', String(5)),
                  Column('alpha', String(5)),
                  Column('beta', String(5)),
                  Column('gamma', String(5)),
                  Column('cif', String(25)), ## , nullable=True
                  Column('zstr',String(25)),
                  Column('qstr',String(25)),
                  Column('url',String(25))
                  )
        ###################################################
        ## Add all to file
        self.metadata.create_all() ## if not exists function (callable when exists)

        ###################################################
        ## Define 'add/insert' functions for each table
        def_elem = elemtbl.insert()
        def_name = nametbl.insert()
        def_form = formtbl.insert()
        def_spgp = spgptbl.insert()
        def_sym  = symtbl.insert()
        def_auth = authtbl.insert()
        def_q    = qtbl.insert()
        def_cat  = cattbl.insert()

        add_sym  = symref.insert()
        add_comp = compref.insert()
        add_q    = qref.insert()
        add_auth = authref.insert()
        add_cat  = catref.insert()

        new_cif  = ciftbl.insert()


        ###################################################
        ## Populate the fixed tables of the database

        ## Adds all elements into database
        for element in ELEMENTS:
            z, name, symbol = element
            def_elem.execute(z=int(z), element_name=name, element_symbol=symbol)

        ## Adds all crystal symmetries
        for symmetry_id,symmetry in enumerate(SYMMETRIES):
            def_sym.execute(symmetry_name=symmetry.strip())
            if symmetry.strip() == 'triclinic':      ## triclinic    :   1 -   2
                for iuc_id in range(1,2+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))
            elif symmetry.strip() == 'monoclinic':   ## monoclinic   :   3 -  15
                for iuc_id in range(3,15+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))
            elif symmetry.strip() == 'orthorhombic': ## orthorhombic :  16 -  74
                for iuc_id in range(16,74+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))
            elif symmetry.strip() == 'tetragonal':   ## tetragonal   :  75 - 142
                for iuc_id in range(75,142+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))
            elif symmetry.strip() == 'trigonal':     ## trigonal     : 143 - 167
                for iuc_id in range(143,167+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))
            elif symmetry.strip() == 'hexagonal':    ## hexagonal    : 168 - 194
                for iuc_id in range(168,194+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))
            elif symmetry.strip() == 'cubic':        ## cubic        : 195 - 230
                for iuc_id in range(195,230+1):
                    add_sym.execute(iuc_id=iuc_id,symmetry_id=(symmetry_id+1))

        for cat in CATEGORIES:
            def_cat.execute(category_name=cat)

        ## Adds qrange
        for q in QAXIS:
            def_q.execute(q='%0.2f' % q)

        ## Adds all space groups
        for spgrp_no in SPACEGROUPS.keys():
            for spgrp_name in SPACEGROUPS[spgrp_no]:
                try:
                    def_spgp.execute(iuc_id=spgrp_no, hm_notation=spgrp_name)
                except:
                    if verbose:
                        print('Duplicate: %s %s' % (spgrp_no,spgrp_name))
                pass


    def __add_space_groups(self):
        ## Add missing space groups
        for spgrp_no in SPACEGROUPS.keys():
            for spgrp_name in SPACEGROUPS[spgrp_no]:
                match = False
                search_spgrp = self.spgptbl.select(self.spgptbl.c.hm_notation == spgrp_name)
                for row in search_spgrp.execute():
                    match = True
                if match is False:
                    print('Adding: %s %s' % (spgrp_no,spgrp_name))
                    self.spgptbl.insert().execute(iuc_id=spgrp_no,hm_notation=spgrp_name)

    def add_ciffile(self, ciffile, verbose=True, url=False, ijklm=1, file=None):
        '''
        ## Adds ciffile into database
        When reading in new CIF:
        i.   put entire cif into field
        ii.  read _database_code_amcsd; write 'amcsd_id' to 'cif data'
        iii. read _chemical_name_mineral; find/add in' minerallist'; write
             'mineral_id' to 'cif data'
        iv.  read _symmetry_space_group_name_H-M - find in 'spacegroup'; write
             iuc_id to 'cif data'
        v.   read author name(s) - find/add in 'authorlist'; write 'author_id',
             'amcsd_id' to 'authref'
        vi.  read _chemical_formula_sum; write 'z' (atomic no.), 'amcsd_id'
             to 'compref'
        vii. calculate q - find each corresponding 'q_id' for all peaks; in write
            'q_id','amcsd_id' to 'qpeak'
        '''

        if url:
            cifstr = requests.get(ciffile).text
        else:
            with open(ciffile,'rb') as file:
                cifstr = str(file.read().decode('utf-8'))
        cif = create_xrdcif(text=cifstr)

        if cif.id_no is None:
            cif_no = 99999
            search_cif = self.query(self.ciftbl.c.amcsd_id).filter(self.ciftbl.c.amcsd_id == cif_no).all()
            cnt_lp = 0
            while len(search_cif) > 0:
                cif_no += 1
                cnt_lp += 1
                search_cif = self.query(self.ciftbl.c.amcsd_id).filter(self.ciftbl.c.amcsd_id == cif_no).all()
                if cnt_lp > 500: ##  safe guards against infinite loop
                    print(' *** too many loops to find unassigned AMCSD number.')
                    return
            print(' *** Assigning unnumbered CIF to AMCSD %i' % cif_no)
            cif.id_no = cif_no

        ## check for amcsd in file already
        ## Find amcsd_id in database
        self.ciftbl = Table('ciftbl', self.metadata)
        search_cif = self.ciftbl.select(self.ciftbl.c.amcsd_id == cif.id_no)
        for row in search_cif.execute():
            if verbose:
                if url:
                    print('AMCSD %i already exists in database.\n' % cif.id_no)
                else:
                    print('%s: AMCSD %i already exists in database %s.' %
                         (os.path.split(ciffile)[-1],cif.id_no,self.dbname))
            return

        ## Define q-array for each entry at given energy
        qhkl = cif.calc_q(wvlgth=lambda_from_E(ENERGY), q_min=QMIN, q_max=QMAX)
        qarr = self.create_q_array(qhkl)

        ###################################################
        def_name = self.nametbl.insert()
        def_form = self.formtbl.insert()
        def_spgp = self.spgptbl.insert()
        def_sym  = self.symtbl.insert()
        def_auth = self.authtbl.insert()
        def_q    = self.qtbl.insert()
        def_cat  = self.cattbl.insert()
        add_sym  = self.symref.insert()
        add_comp = self.compref.insert()
        add_q    = self.qref.insert()
        add_auth = self.authref.insert()
        add_cat  = self.catref.insert()
        new_cif  = self.ciftbl.insert()

        ## Find mineral_name
        match = False
        search_mineral = self.nametbl.select(self.nametbl.c.mineral_name == cif.label)
        for row in search_mineral.execute():
            mineral_id = row.mineral_id
            match = True
        if match is False:
            def_name.execute(mineral_name=cif.label)
            search_mineral = self.nametbl.select(self.nametbl.c.mineral_name == cif.label)
            for row in search_mineral.execute():
                mineral_id = row.mineral_id

        ## Find formula_name
        match = False
        search_formula = self.formtbl.select(self.formtbl.c.formula_name == cif.formula)
        for row in search_formula.execute():
            formula_id = row.formula_id
            match = True
        if match is False:
            def_form.execute(formula_name=cif.formula)
            search_formula = self.formtbl.select(self.formtbl.c.formula_name == cif.formula)
            for row in search_formula.execute():
                formula_id = row.formula_id

        ## Find composition (loop over all elements)
        z_list = []
        for element in set(cif.atom.label):
            search_elements = self.elemtbl.select(self.elemtbl.c.element_symbol == element)
            for row in search_elements.execute():
                z_list += [row.z]
        zarr = self.create_z_array(z_list)


        ## Save CIF entry into database
        new_cif.execute(amcsd_id=cif.id_no,
                        mineral_id=int(mineral_id),
                        formula_id=int(formula_id),
                        iuc_id=cif.symmetry.no,
                        a=str(cif.unitcell[0]),
                        b=str(cif.unitcell[1]),
                        c=str(cif.unitcell[2]),
                        alpha=str(cif.unitcell[3]),
                        beta=str(cif.unitcell[4]),
                        gamma=str(cif.unitcell[5]),
                        cif=cifstr,
                        zstr=json.dumps(zarr.tolist(),default=str),
                        qstr=json.dumps(qarr.tolist(),default=str),
                        url=str(ciffile))

        ## Build q cross-reference table
        for q in qhkl:
            search_q = self.qtbl.select(self.qtbl.c.q == '%0.2f' % (int(q * 100) / 100.))
            for row in search_q.execute():
                q_id = row.q_id

            try:
                add_q.execute(q_id=q_id,amcsd_id=cif.id_no)
            except:
                pass


        ## Build composition cross-reference table
        for element in set(cif.atom.label):
            search_elements = self.elemtbl.select(self.elemtbl.c.element_symbol == element)
            for row in search_elements.execute():
                z = row.z

            try:
                add_comp.execute(z=z, amcsd_id=cif.id_no)
            except:
                print('could not find element: %s (amcsd: %i)' % (element,cif.id_no))
                pass

        ## Find author_name
        for author_name in cif.publication.author:
            match = False
            search_author = self.authtbl.select(self.authtbl.c.author_name == author_name)
            for row in search_author.execute():
                author_id = row.author_id
                match = True
            if match is False:
                def_auth.execute(author_name=author_name)
                search_author = self.authtbl.select(self.authtbl.c.author_name == author_name)
                for row in search_author.execute():
                    author_id = row.author_id
                    match = True
            if match == True:
                add_auth.execute(author_id=author_id,
                                   amcsd_id=cif.id_no)

    #     ## not ready for defined categories
    #     cif_category.execute(category_id='none',
    #                          amcsd_id=cif.id_no)

        if url:
            self.amcsd_info(cif.id_no, no_qpeaks=np.sum(qarr))
        else:
            self.amcsd_info(cif.id_no, no_qpeaks=np.sum(qarr),ciffile=ciffile)

    def url_to_cif(self, url=None, verbose=False, savecif=False, addDB=True,
                   all=False, minval=None):

        maxi = 20573
        exceptions = [0,7271,10783,14748,15049,15050,15851,18368,
                      18449,18450,18451,18452,18453,20029]

        ## ALL CAUSE FAILURE IN CIFFILE FUNCTION:
        ##  7271 : author name doubled in cif
        ## 14748 : has label of amcsd code but no number (or anything) assigned
        ## 15049 : page number 'L24307 1' could not be parsed as number
        ## 15050 : page number 'L24307 1' could not be parsed as number
        ## 15851 : no first page number provided despite providing field label
        ## 18368 : non-numerical entries in B_iso fields
        ## 18449 : no first page number provided despite providing field label
        ## 18450 : no first page number provided despite providing field label
        ## 20029 : no volume number provided despite providing field label

        if url is None:
            url = 'http://rruff.geo.arizona.edu/AMS/download.php?id=%05d.cif&down=cif'

        ## Defines url range for searching and adding to cif database
        if all:
            iindex = range(99999) ## trolls whole database online
        elif minval is not None:
            iindex = np.arange(minval, 99999) ## starts at given min and counts up
        else:
            iindex = np.arange(13600, 13700) ## specifies small range including CeO2 match

        for i in iindex:
            if i not in exceptions and i < maxi:
                url_to_scrape = url % i
                r = requests.get(url_to_scrape)
                if r.text.split()[0] == "Can't" or '':
                    if verbose:
                        print('\t---> ERROR on amcsd%05d.cif' % i)
                else:
                    if verbose:
                        print('Reading %s' % url_to_scrape)
                    if savecif:
                        file = 'amcsd%05d.cif' % i
                        f = open(file,'w')
                        f.write(r.text)
                        f.close()
                        if verbose:
                            print('Saved %s' % file)
                    if addDB:
                        try:
                            self.add_ciffile(url_to_scrape, url=True, verbose=verbose, ijklm=i)
                        except:
                            pass




##################################################################################
##################################################################################

#         usr_qry = self.query(self.ciftbl,
#                              self.elemtbl,self.nametbl,self.spgptbl,self.symtbl,
#                              self.authtbl,self.qtbl,self.cattbl,
#                              self.authref,self.compref,self.catref,self.symref)\
#                       .filter(self.authref.c.amcsd_id == self.ciftbl.c.amcsd_id)\
#                       .filter(self.authtbl.c.author_id == self.authref.c.author_id)\
#                       .filter(self.compref.c.amcsd_id == self.ciftbl.c.amcsd_id)\
#                       .filter(self.compref.c.z == self.elemtbl.c.z)\
#                       .filter(self.catref.c.amcsd_id == self.ciftbl.c.amcsd_id)\
#                       .filter(self.catref.c.category_id == self.cattbl.c.category_id)\
#                       .filter(self.nametbl.c.mineral_id == self.ciftbl.c.mineral_id)\
#                       .filter(self.symref.c.symmetry_id == self.symtbl.c.symmetry_id)\
#                       .filter(self.symref.c.iuc_id == self.spgptbl.c.iuc_id)\
#                       .filter(self.spgptbl.c.iuc_id == self.ciftbl.c.iuc_id)

##################################################################################
##################################################################################


##################################################################################

    def amcsd_info(self, amcsd_id, no_qpeaks=None, ciffile=None):
        mineral_id,iuc_id = self.cif_by_amcsd(amcsd_id,only_ids=True)
        mineral_name = self.search_for_mineral(minid=mineral_id)[0].mineral_name
        authors      = self.author_by_amcsd(amcsd_id)

        ## ALLelements,mineral_name,iuc_id,authors = self.all_by_amcsd(amcsd_id)

        if ciffile:
            print(' ==== File : %s ====' % os.path.split(ciffile)[-1])
        else:
            print(' ===================== ')
        print(' AMCSD: %i' % amcsd_id)
        print(' Name: %s' % mineral_name)
        print(' %s' % self.composition_by_amcsd(amcsd_id))
        try:
            print(' Space Group No.: %s (%s)' % (iuc_id,self.symm_id(iuc_id)))
        except:
            print(' Space Group No.: %s' % iuc_id)
        if no_qpeaks:
            print(' No. q-peaks in range : %s' % no_qpeaks)

        authorstr = ' Author(s): '
        for author in authors:
             authorstr = '%s %s' % (authorstr,author.split()[0])
        print(authorstr)
        print(' ===================== ')

    def symm_id(sel, iuc_id):

        if not isinstance(iuc_id, int):
            iuc_id = int(iuc_id.split(':')[0])

        if   iuc_id < 3  : return 'triclinic'    ##   1 -   2 : Triclinic
        elif iuc_id < 16 : return 'monoclinic'   ##   3 -  15 : Monoclinic
        elif iuc_id < 75 : return 'orthorhombic' ##  16 -  74 : Orthorhombic
        elif iuc_id < 143: return 'tetragonal'   ##  75 - 142 : Tetragonal
        elif iuc_id < 168: return 'trigonal'     ## 143 - 167 : Trigonal
        elif iuc_id < 195: return 'hexagonal'    ## 168 - 194 : Hexagonal
        elif iuc_id < 231: return 'cubic'        ## 195 - 230 : Cubic
        else:
            return

    def return_cif(self,amcsd_id):
        search_cif = self.ciftbl.select(self.ciftbl.c.amcsd_id == amcsd_id)
        for row in search_cif.execute():
            return row.cif

##################################################################################

    def all_by_amcsd(self, amcsd_id):

        mineral_id,iuc_id = self.cif_by_amcsd(amcsd_id,only_ids=True)

        mineral_name = self.search_for_mineral(minid=mineral_id)[0].mineral_name
        ALLelements  = self.composition_by_amcsd(amcsd_id)
        authors      = self.author_by_amcsd(amcsd_id)

        return ALLelements, mineral_name, iuc_id, authors

    def q_by_amcsd(self,amcsd_id,qmin=QMIN,qmax=QMAX):

        q_results = self.query(self.ciftbl.c.qstr).filter(self.ciftbl.c.amcsd_id == amcsd_id).all()
        q_all = [json.loads(qrow[0]) for qrow in q_results]

        return [self.axis[i] for i,qi in enumerate(q_all[0]) if qi == 1 and self.axis[i] >= qmin and self.axis[i] <= qmax]

    def author_by_amcsd(self,amcsd_id):

        search_authors = self.authref.select(self.authref.c.amcsd_id == amcsd_id)
        authors = []
        for row in search_authors.execute():
            authors.append(self.search_for_author(row.author_id,id_no=False)[0][0])
        return authors

    def composition_by_amcsd(self, amcsd_id):
        q = self.query(self.compref).filter(self.compref.c.amcsd_id==amcsd_id)
        return [row.z for row in q.all()]

    def cif_by_amcsd(self,amcsd_id,only_ids=False):

        search_cif = self.ciftbl.select(self.ciftbl.c.amcsd_id == amcsd_id)
        for row in search_cif.execute():
            if only_ids:
                return row.mineral_id, row.iuc_id
            else:
                return row.cif

    def mineral_by_amcsd(self,amcsd_id):

        search_cif = self.ciftbl.select(self.ciftbl.c.amcsd_id == amcsd_id)
        for row in search_cif.execute():
            cifstr = row.cif
            mineral_id = row.mineral_id
            iuc_id = row.iuc_id

        search_mineralname = self.nametbl.select(self.nametbl.c.mineral_id == mineral_id)
        for row in search_mineralname.execute():
            mineral_name = row.mineral_name
        return mineral_name

##################################################################################
##################################################################################

    def amcsd_by_q(self, peaks, qmin=None, qmax=None, qstep=None, list=None,
                   verbose=False):

        if qmin is None: qmin = QMIN
        if qmax is None: qmax = QMAX
        if qstep is None: qstep = QSTEP

        ## Defines min/max limits of q-range
        imin, imax = 0, len(self.axis)
        if qmax < np.max(self.axis):
            imax = abs(self.axis-qmax).argmin()
        if qmin > np.min(self.axis):
            imin = abs(self.axis-qmin).argmin()
        qaxis = self.axis[imin:imax]
        stepq = (qaxis[1]-qaxis[0])

        amcsd, q_amcsd = self.match_qc(list=list, qmin=qmin, qmax=qmax)

        ## Re-bins data if different step size is specified
        if qstep > stepq:
            new_qaxis   = np.arange(np.min(qaxis),np.max(qaxis)+stepq,qstep)
            new_q_amcsd = np.zeros((np.shape(q_amcsd)[0],np.shape(new_qaxis)[0]))
            for m,qrow in enumerate(q_amcsd):
                for n,qn in enumerate(qrow):
                    if qn == 1:
                        k = np.abs(new_qaxis-qaxis[n]).argmin()
                        new_q_amcsd[m][k] = 1
            qaxis = new_qaxis
            q_amcsd = new_q_amcsd


        ## Create data array
        peaks_weighting = np.ones(len(qaxis),dtype=int)*-1
        peaks_true = np.zeros(len(qaxis),dtype=int)
        peaks_false = np.ones(len(qaxis),dtype=int)
        for p in peaks:
            i = np.abs(qaxis-p).argmin()
            peaks_weighting[i],peaks_true[i],peaks_false[i] = 1,1,0

        ## Calculate score/matches/etc.
        total_peaks = np.sum((q_amcsd),axis=1)
        match_peaks = np.sum((peaks_true*q_amcsd),axis=1)
        miss_peaks = np.sum((peaks_false*q_amcsd),axis=1)
        scores = np.sum((peaks_weighting*q_amcsd),axis=1)

        return sorted(zip(scores, amcsd, total_peaks, match_peaks, miss_peaks), reverse=True)


    def amcsd_by_chemistry(self, include=[], exclude=[]):

        amcsd_incld = []
        amcsd_excld = []
        z_incld = []
        z_excld = []

        if len(include) > 0:
            for element in include:
                z = self.get_element(element).z
                if z is not None and z not in z_incld:
                    z_incld += [z]
        if isinstance(exclude,bool):
            if exclude:
                for element in ELEMENTS:
                    z, name, symbol = element
                    z = int(z)
                    if z not in z_incld:
                        z_excld += [z]
        else:
            if len(exclude) > 0:
                for element in exclude:
                    z = self.get_element(element).z
                    if z is not None and z not in z_excld:
                        z_excld += [z]

        z_list_include = [1 if z in z_incld else 0 for z in np.arange(len(ELEMENTS)+1)]
        z_list_exclude = [1 if z in z_excld else 0 for z in np.arange(len(ELEMENTS)+1)]

        amcsd,z_amcsd = self.return_z_matches(list=list)

        ## Calculate score/matches/etc.
        match_z = np.sum((np.array(z_list_include)*np.array(z_amcsd)),axis=1)
        miss_z  = np.sum((np.array(z_list_exclude)*np.array(z_amcsd)),axis=1)

        for i,amcsd_id in enumerate(amcsd):
            if match_z[i] == np.sum(z_list_include) and miss_z[i] <= 0:
                amcsd_incld += [amcsd_id]
            else:
                amcsd_excld += [amcsd_id]
        return amcsd_incld


    def amcsd_by_mineral(self, min_name, list=None, verbose=True):
        """
        search by mineral name
        """
        out = []
        minerals = self.search_for_mineral(name=min_name)

        q = self.query(self.ciftbl)
        if list is not None:
            q = q.filter(self.ciftbl.c.amcsd_id.in_(list))

        ##  Searches mineral name for database entries
        if len(minerals) > 0:
            mids = [m.mineral_id for m in minerals]
            q = q.filter(self.ciftbl.c.mineral_id.in_(mids))
            for row in q.all():
                if row.amcsd_id not in out:
                    out.append(row.amcsd_id)
        return out

    def amcsd_by_author(self,include=[''],list=None,verbose=True):

        amcsd_incld = []
        auth_id = []

        for author in include:
            id = self.search_for_author(author)
            auth_id += id

        ##  Searches mineral name for database entries
        usr_qry = self.query(self.ciftbl,self.authtbl,self.authref)\
                      .filter(self.authref.c.amcsd_id == self.ciftbl.c.amcsd_id)\
                      .filter(self.authref.c.author_id == self.authtbl.c.author_id)
        if list is not None:
            usr_qry = usr_qry.filter(self.ciftbl.c.amcsd_id.in_(list))

        ##  Searches author name in database entries
        if len(auth_id) > 0:
            fnl_qry = usr_qry.filter(self.authref.c.author_id.in_(auth_id))
            ## This currently works in an 'or' fashion, as each name in list
            ## can be matched to multiple auth_id values, so it is simpler to
            ## consider them all separately. Making a 2D list and restructuring
            ## query could improve this
            ## mkak 2017.02.24
            for row in fnl_qry.all():
                if row.amcsd_id not in amcsd_incld:
                    amcsd_incld += [row.amcsd_id]

        return amcsd_incld


    def match_elements(self, elems, exclude=None):
        """match structues containing all elements in a list

        Arguments:
        ----------
        elems    list of elements to match
        exclude  list of elements to exclude for match (default None)

        Returns:
        --------
        list of amcsd ids for structures

        """
        matches = None
        q = self.query(self.compref)

        for elem in elems:
            elem = self.get_element(elem).z
            rows = q.filter(self.compref.c.z==elem).all()
            sids = [row.amcsd_id for row in rows]
            if matches is None:
                matches = sids
            else:
                matches = [s for s in sids if s in matches]

        if exclude is not None:
            for elem in exclude:
                elem = self.get_element(elem).z
                for row in q.filter(self.compref.c.z==elem).all():
                    if row.amcsd_id in matches:
                        matches.remove(row.amcsd_id)
        return matches

    def create_z_array(self,z):
        z_array = np.zeros((len(ELEMENTS)+1),dtype=int) ## + 1 gives index equal to z; z[0]:nothing
        for zn in z:
            z_array[zn] = 1
        return z_array


##################################################################################
##################################################################################
    def match_qc(self, list=None, qmin=QMIN, qmax=QMAX):

        if list is None:
            qqry = self.query(self.ciftbl.c.qstr).all()
            idqry = self.query(self.ciftbl.c.amcsd_id).all()
        else:
            qqry = self.query(self.ciftbl.c.qstr)\
                   .filter(self.ciftbl.c.amcsd_id.in_(list))\
                   .all()
            idqry = self.query(self.ciftbl.c.amcsd_id)\
                    .filter(self.ciftbl.c.amcsd_id.in_(list))\
                    .all()

        imin,imax = 0,len(self.axis)
        if qmax < QMAX: imax = abs(self.axis-qmax).argmin()
        if qmin > QMIN: imin = abs(self.axis-qmin).argmin()

        return [id[0] for id in idqry],[json.loads(q[0])[imin:imax] for q in qqry]

    def create_q_array(self, q):

        q_array = np.zeros(len(self.axis), dtype=int)
        for qn in q:
            i = np.abs(self.axis-qn).argmin()
            q_array[i] = 1
        return q_array

##################################################################################

    def get_element(self, element):
        '''
        searches elements for match in symbol, name, or atomic number;
        match must be exact.

        returns row with attributes  .z, .element_name, .element_symbol
        '''
        if isinstance(element, int):
            element = '%d' % element
        elif isinstance(element, str):
            element = element.title()
        q = self.query(self.elemtbl)
        row = q.filter(or_(self.elemtbl.c.z == element,
                           self.elemtbl.c.element_symbol == element,
                           self.elemtbl.c.element_name == element)).one()
        return row

    def search_for_author(self,name,exact=False,id_no=True,verbose=False):
        '''
        searches database for author matching criteria given in 'name'
           - if name is a string:
                  - will match author name containing text
                  - will match id number if integer given in string
                  - will only look for exact match if exact flag is given
           - if name is an integer, will only match id number from database
        id_no: if True, will only return the id number of match(es)
               if False, returns name and id number
        e.g.   as INTEGER
               >>> cif.search_for_author(6,id_no=False)
                    ([u'Chao G Y'], [6])
               as STRING
               >>> cif.search_for_author('6',id_no=False)
                    ([u'Chao G Y', u'Geology Team 654'], [6, 7770])
        '''

        authname = []
        authid   = []

        id, name = filter_int_and_str(name,exact=exact)
        authrow = self.query(self.authtbl)\
                      .filter(or_(self.authtbl.c.author_name.like(name),
                                  self.authtbl.c.author_id  == id))
        if len(authrow.all()) == 0:
            if verbose: print('%s not found in author database.' % name)
        else:
            for row in authrow.all():
                authname += [row.author_name]
                authid   += [row.author_id]

        if id_no: return authid
        else: return authname,authid

    def search_for_mineral(self, name=None, minid=None, exact=False):
        '''
        searches database for mineral by name or by ID

        Arguments:
        ----------
          name  (str or None): mineral name to match
          minid (int or None): mineral ID in database to match
          exact (bool):  whether to match name exactly [False]


        Returns:
        --------
        list of matching rows

        # [row.mineral_name, row.mineral_id]
        '''

        rows = []
        q = self.query(self.nametbl)

        if name is not None:
            if not exact:
                name = '%%%s%%' % name
            rows = q.filter(self.nametbl.c.mineral_name.like(name)).all()
        elif minid is not None:
            rows = q.filter(self.nametbl.c.mineral_id == minid).all()
        return rows

    def cif_count(self):
        return self.query(self.ciftbl).count()

    def return_q(self):
        q = [float(row.q) for row in self.query(self.qtbl).all()]
        return np.array(q)

    def get_mineral_names(self):
        names = []
        for name in self.query(self.nametbl.c.mineral_name).all():
            if isinstance(name[0], str):
                names.append(name[0])
        return sorted(names)

    def return_author_names(self):

        authorqry = self.query(self.authtbl)
        names = []
        for row in authorqry.all():
            names += [row.author_name]

        return sorted(names)

def filter_int_and_str(s, exact=False):
        try:
            i = int(s)
        except:
            i = 0
        if not exact:
            try:
                s = '%'+s+'%'
            except:
                pass
        return i, s


def column(matrix, i):
    return [row[i] for row in matrix]

class RangeParameter(object):
    def __init__(self,min=None,max=None,unit=None):
        self.min   = min
        self.max   = max
        self.unit  = unit

class SearchCIFdb(object):
    '''
    interface to the search the cif database
    '''
    def __init__(self, verbose=False):

        self.verbose = verbose

        ## running list of included amcsd id numbers
        self.amcsd_id = []

        ## tags for searching
        self.authors    = []
        self.keywords   = []
        self.categories = []
        self.amcsd      = []
        self.qpks       = []

        self.mnrlname   = ''

        self.elem_incl = []
        self.elem_excl = []
        self.allelem   = column(ELEMENTS, 2)

        self.lattice_keys = ['a', 'b', 'c', 'alpha', 'beta', 'gamma']

        self.sg    = None
        self.a     = RangeParameter()
        self.b     = RangeParameter()
        self.c     = RangeParameter()
        self.alpha = RangeParameter()
        self.beta  = RangeParameter()
        self.gamma = RangeParameter()


    def show_all(self):
        for key in ['authors','mnrlname','keywords','categories','amcsd','qpks']:
             print('%s : %s' % (key,self.show_parameter(key=key)))
        print('chemistry : %s' % self.show_chemistry())
        print('geometry : %s' % self.show_geometry())

    def show_parameter(self, key='authors'):
        s = ''
        if len(self.__dict__[key]) > 0:
            for i,item in enumerate(self.__dict__[key]):
                item = item.split()[0]
                if i == 0:
                    s = '%s' % (item)
                else:
                    s = '%s, %s' % (s,item)
        return s


    def read_parameter(self,s,clear=True,key='authors'):
        '''
        This function works for keys:
        'authors'
        'mnrlname
        keywords','categories','amcsd','qpks'
        '''

        if clear:
            self.__dict__[key] = []
        if len(s) > 0:
            for a in s.split(','):
                try:
                    self.__dict__[key] += [a.split()[0]]
                except:
                    pass

    def read_chemistry(self,s,clear=True):

        if clear:
            self.elem_incl,self.elem_excl = [],[]
        chem_incl,chem_excl = [],[]

        chemstr = re.sub('[( )]','',s)
        ii = -1
        for i,s in enumerate(chemstr):
            if s == '-':
                ii = i
        if ii > 0:
            chem_incl = chemstr[0:ii].split(',')
            if len(chemstr)-ii == 1:
                for elem in self.allelem:
                    if elem not in chem_incl:
                        chem_excl += [elem]
            elif ii < len(chemstr)-1:
                chem_excl = chemstr[ii+1:].split(',')
        else:
            chem_incl = chemstr.split(',')

        for elem in chem_incl:
            elem = capitalize_string(elem)
            if elem in self.allelem and elem not in self.elem_incl:
                self.elem_incl += [elem]
                if elem in self.elem_excl:
                    j = self.elem_excl.index(elem)
                    self.elem_excl.pop(j)
        for elem in chem_excl:
            elem = capitalize_string(elem)
            if elem in self.allelem and elem not in self.elem_excl and elem not in self.elem_incl:
                self.elem_excl += [elem]

    def show_chemistry(self):

        s = ''
        for i,elem in enumerate(self.elem_incl):
            if i==0:
                s = '(%s' % elem
            else:
                s = '%s,%s' % (s,elem)
        if len(self.elem_incl) > 0:
            s = '%s) ' % s
        if len(self.elem_excl) > 0:
            s = '%s- ' % s
        # if all else excluded, don't list
        if (len(self.allelem)-20) > (len(self.elem_incl)+len(self.elem_excl)):
            for i,elem in enumerate(self.elem_excl):
                if i==0:
                    s = '%s(%s' % (s,elem)
                else:
                    s = '%s,%s' % (s,elem)
            if len(self.elem_excl) > 0:
                s = '%s)' % s
        return s

    def show_geometry(self,unit='A'):

        s = ''

        key = 'sg'
        if self.__dict__[key] is not None:
            s = '%s%s=%s,' % (s,key,self.__dict__[key])
        for i,key in enumerate(self.lattice_keys):
            if self.__dict__[key].min is not None:
                s = '%s%s=%0.2f' % (s,key,float(self.__dict__[key].min))
                if self.__dict__[key].max is not None:
                    s = '%sto%0.2f' % (s,float(self.__dict__[key].max))
                s = '%s%s,' % (s,self.__dict__[key].unit)

        if len(s) > 1:
            if s[-1] == ',':
                s = s[:-1]

        return s

    def read_geometry(self,s):

        geostr = s.split(',')
        used = []
        for par in geostr:
            key = par.split('=')[0]
            val = par.split('=')[1]
            if key in 'sg':
                self.__dict__[key] = val
                used += [key]
            elif key in self.lattice_keys:
                values = [''.join(g) for _, g in groupby(val, str.isalpha)]
                self.__dict__[key].min = values[0]
                if len(values) > 1:
                    self.__dict__[key].unit = values[-1]
                if len(values) > 2:
                    self.__dict__[key].max = values[2]
                else:
                    self.__dict__[key].max = None
                used += [key]

        ## Resets undefined to None
        for key in self.lattice_keys:
            if key not in used:
                self.__dict__[key] = RangeParameter()
        key = 'sg'
        if key not in used:
            self.__dict__[key] = None

def match_database(cifdb, peaks, minq=QMIN, maxq=QMAX, verbose=True):
    """
    fracq  : min. ratio of matched q to possible in q range, i.e. 'goodness gauge'
    pk_wid : maximum range in q which qualifies as a match between fitted and ideal
    """
    stepq = 0.05
    scores,amcsd,total_peaks,match_peaks,miss_peaks = zip(*cifdb.amcsd_by_q(peaks,
                                                       qmin=minq,qmax=maxq,qstep=stepq,
                                                       list=None,verbose=False))

    MATCHES = [match for i,match in enumerate(amcsd) if scores[i] > 0]

    if verbose:
        print('\n')
        if len(MATCHES) > 100:
            print('DISPLAYING TOP 100 of %i TOTAL MATCHES FOUND.' % len(MATCHES))
        else:
            print('%i TOTAL MATCHES FOUND.' % len(MATCHES))
        j = 0
        for i,id_no in enumerate(amcsd):
            if j < 100:
                if scores[i] > 0:
                    j += 1
                    str = 'AMCSD %5d, %s (score of %2d --> %i of %i peaks)' % (id_no,
                             cifdb.mineral_by_amcsd(id_no),scores[i],
                             match_peaks[i],total_peaks[i])
                    print(str)
        print('')

    return MATCHES


def cif_match(peaks, qmin=None, qmax=None, verbose=False, _larch=None):
    """
    fracq  : min. ratio of matched q to possible in q range, i.e. 'goodness gauge'
    pk_wid : maximum range in q which qualifies as a match between fitted and ideal
    """
    cifdb = get_cifdb(_larch=_larch)
    qstep = 0.05

    rows = cifdb.amcsd_by_q(peaks, qmin=qmin,qmax=qmax, qstep=qstep)

    scores, amcsd, total_peaks, match_peaks, miss_peaks = rows

    matches = []
    for i, cdat in enumerate(amcsd):
        if score[i] > 0:
            matches.append(cdat)

    if verbose:
        print('\n')
        if len(MATCHES) > 100:
            print('DISPLAYING TOP 100 of %i TOTAL MATCHES FOUND.' % len(MATCHES))
        else:
            print('%i TOTAL MATCHES FOUND.' % len(MATCHES))
        matches = matches[:100]
        for i, id_no in enumerate(amcsd):
            if j < 100:
                if scores[i] > 0:
                    j += 1
                    str = 'AMCSD %5d, %s (score of %2d --> %i of %i peaks)' % (id_no,
                             cifdb.mineral_by_amcsd(id_no),scores[i],
                             match_peaks[i],total_peaks[i])
                    print(str)
        print('')

    return matches


def read_cif(filename=None, amcsd_id=None, _larch=None):
    """make a representation of a CIF data structure
    for crystallographic computations

    Arguments:
    ----------
    filename  (str or None) name of CIF file
    amcsd_id  (int or None) index of CIF in Am Min Cystal Structure database

    Returns
    -------
    CIF representation
    """
    cifdb = get_cifdb(_larch=_larch)
    return create_xrdcif(filename=filename, cifdb=cifdb, amcsd_id=amcsd_id)



import numpy.testing
from larch.math import utils

tiny=1.e-6

def test_remove_dups():
    expected = numpy.array([])
    calculated = utils.remove_dups([])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([1])
    calculated = utils.remove_dups([1])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([1, 1 + tiny])
    calculated = utils.remove_dups([1, 1])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([1, 1 + tiny, 1 + 2*tiny])
    calculated = utils.remove_dups([1, 1, 1])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([1, numpy.nan])
    calculated = utils.remove_dups([1, numpy.nan])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([1, numpy.nan, 1 + tiny])
    calculated = utils.remove_dups([1, numpy.nan, 1])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([numpy.nan, 1, 1 + tiny])
    calculated = utils.remove_dups([numpy.nan, 1, 1])
    numpy.testing.assert_array_equal(calculated, expected)

    expected = numpy.array([numpy.nan, 1, 1 + tiny, 1 + 2*tiny])
    calculated = utils.remove_dups([[numpy.nan, 1], [1, 1]])
    numpy.testing.assert_allclose(calculated, expected, atol=tiny/4)



from larch.io import read_athena
from glob import glob

passed = 0
for fname in sorted(glob('*.prj')):
    print("### ", fname)
    group = read_athena(fname)
    print('read ', group, list(group.groups.keys()))
    passed += 1


print(f"Read {passed:d} Athena Project files")



import numpy as np
import pylab
from larch import Interpreter, Group

# now import larch-specific Python code
from larch_plugins.xafs import autobk, xftf

# create a larch interpreter, passed to most functions
my_larch = Interpreter(with_plugins=False)

# create an empty Group
xafsdat = Group()

# read data (here using np.loadtxt), stick into Group
# using "expected names" for XAFS data
rawdata = np.loadtxt('../xafsdata/fe2o3_rt1.xmu')

xafsdat.energy = rawdata[:, 0]
xafsdat.mu     = rawdata[:, 1]
xafsdat.i0     = rawdata[:, 1]

# run autobk on the xafsdat Group, including a larch Interpreter....
# note that this expects 'energy' and 'mu' to be in xafsdat, and will
# write data for 'k', 'chi', 'kwin', 'e0', ... into xafsdat
autobk(xafsdat, rbkg=1.0, kweight=2, _larch=my_larch)

# Fourier transform to R space, again passing in a Group (here,
# 'k' and 'chi' are expected, and writitng out 'r', 'chir_mag',
# and so on
xftf(xafsdat, kmin=2, kmax=15, dk=3, kweight=2, _larch=my_larch)

#
# plot grid of results:
# mu + bkg
pylab.subplot(2, 2, 1)
pylab.plot(xafsdat.energy, xafsdat.bkg, 'r--')
pylab.plot(xafsdat.energy, xafsdat.mu)
pylab.xlabel('Energy (eV)')

# normalized XANES
# find array bounds for normalized mu(E) for [e0 - 25: e0 + 75]
j0 = np.abs(xafsdat.energy-(xafsdat.e0 - 25.0)).argmin()
j1 = np.abs(xafsdat.energy-(xafsdat.e0 + 75.0)).argmin()

pylab.subplot(2, 2, 2)
pylab.plot(xafsdat.energy[j0:j1], xafsdat.norm[j0:j1])
pylab.xlabel('Energy (eV)')

# chi(k)
pylab.subplot(2, 2, 3)
pylab.plot(xafsdat.k, xafsdat.chi*xafsdat.k**2)
pylab.plot(xafsdat.k, xafsdat.kwin, 'r--')
pylab.xlabel(r'$ k (\AA^{-1}) $')
pylab.ylabel(r'$ k^2 \chi(\AA^{-2}) $')

# chi(R)
pylab.subplot(2, 2, 4)
pylab.plot(xafsdat.r, xafsdat.chir_mag)
pylab.plot(xafsdat.r, xafsdat.chir_re, 'r--')
pylab.xlabel(r'$ R (\AA) $')
pylab.ylabel(r'$ \chi(R) (\AA^{-3}) $')

pylab.show()



#!/usr/bin/env python
""" Larch Tests Version 1 """
import unittest
import time
import ast
import numpy as np
from sys import version_info

from utils import TestCase
from larch import Interpreter

class TestEval(TestCase):
    '''testing of asteval'''
    def test_dict_index(self):
        '''dictionary indexing'''
        self.trytext("a_dict = {'a': 1, 'b': 2, 'c': 3, 'd': 4}")
        self.isTrue("a_dict['a'] == 1")
        self.isTrue("a_dict['d'] == 4")

    def test_list_index(self):
        '''list indexing'''
        self.trytext("a_list = ['a', 'b', 'c', 'd', 'o']")
        self.isTrue("a_list[0] == 'a'")
        self.isTrue("a_list[1] == 'b'")
        self.isTrue("a_list[2] == 'c'")

    def test_tuple_index(self):
        '''tuple indexing'''
        self.trytext("a_tuple = (5, 'a', 'x')")
        self.isTrue("a_tuple[0] == 5")
        self.isTrue("a_tuple[2] == 'x'")

    def test_string_index(self):
        '''string indexing'''
        self.trytext("a_string = 'hello world'")
        self.isTrue("a_string[0] == 'h'")
        self.isTrue("a_string[6] == 'w'")
        self.isTrue("a_string[-1] == 'd'")
        self.isTrue("a_string[-2] == 'l'")

    def test_ndarray_index(self):
        '''nd array indexing'''
        self.trytext("a_ndarray = 5*arange(20)")
        self.isTrue("a_ndarray[2] == 10")
        self.isTrue("a_ndarray[4] == 20")

    def test_ndarrayslice(self):
        '''array slicing'''
        self.trytext("a_ndarray = arange(200).reshape(10, 20)")
        self.isTrue("a_ndarray[1:3,5:7] == array([[25,26], [45,46]])")
        self.trytext("y = arange(20).reshape(4, 5)")
        self.isTrue("y[:,3]  == array([3, 8, 13, 18])")
        self.isTrue("y[...,1]  == array([1, 6, 11, 16])")
        self.trytext("y[...,1] = array([2, 2, 2, 2])")
        self.isTrue("y[1,:] == array([5, 2, 7, 8, 9])")
        # print(self.session.symtable["y"])

    def test_while(self):
        '''while loops'''
        self.trytext("""n=0
while n < 8:
    n += 1
endwhile
""")
        self.isValue('n',  8)

        self.trytext("""
n=0
while n < 8:
    n += 1
    if n > 3:
        break
    endif
else:
    n = -1
endwhile
""")
        self.isValue('n',  4)


        self.trytext("""
n=0
while n < 8:   # test while-else
    n += 1
else:
    n += -1
endwhile
""")
        self.isValue('n',  7)

        self.trytext("""
n=0
while n < 10:
    n += 1
    if n < 3:
        continue
    endif
    n += 1
    print( ' n = ', n)
    if n > 5:
        break
    endif
endwhile
""")
        self.isValue('n',  6)

    def test_assert(self):
        'test assert statements'
        self.trytext('n=6')
        out, err = self.trytext('assert n==6')
        self.assertTrue(len(err) == 0)
        out, err = self.trytext('assert n==7')
        errtype, errmsg = err[0].get_error()
        self.assertTrue(errtype == 'AssertionError')

    def test_for(self):
        '''for loops'''
        self.trytext('''
n=0
for i in arange(10):
    n += i
endfor
''')
        self.isValue('n', 45)

        self.trytext('''
n=0
for i in arange(10):
    n += i
else:
    n = -1
endfor
''')
        self.isValue('n', -1)

        self.trytext('''
n=0
for i in arange(10):
    n += i
    if n > 2:
        break
    endif
else:
    n = -1
endfor
''')
        self.isValue('n', 3)

    def test_if(self):
        '''test if'''
        self.trytext("""zero = 0
if zero == 0:
    x = 1
endif
if zero != 100:
    x = x+1
endif
if zero > 2:
    x = x + 1
else:
    y = 33
endif
""")
        self.isValue('x',  2)
        self.isValue('y', 33)

    def test_nestedif(self):
        """test of single-line ifs in if blocks"""
        self.trytext("""
def xtest(xt, out=None):
    if out is None:
        out = 'A'
        if xt < 9:  out = 'B'
        if xt < 5:  out = 'C'
        if xt < 2:  out = 'D'
    #endif
    return out
#enddef

a = xtest(100)
b = xtest(8)
c = xtest(4)
d = xtest(1)
q = xtest(4, out='Q')
""")
        self.isValue('a', 'A')
        self.isValue('b', 'B')
        self.isValue('c', 'C')
        self.isValue('d', 'D')
        self.isValue('q', 'Q')


    def test_print(self):
        '''print (ints, str, ....)'''
        out, err  = self.trytext("print(31)")
        self.assertTrue(out== '31\n')

        out, err = self.trytext("print('%s = %.3f' % ('a', 1.2012345))")
        self.assertTrue(out== 'a = 1.201\n')

        out, err = self.trytext("print('{0:s} = {1:.2f}'.format('a', 1.2012345))")
        self.assertTrue(out== 'a = 1.20\n')

    def test_repr(self):
        '''repr of dict, list'''
        self.trytext("x = {'a': 1, 'b': 2, 'c': 3}")
        self.trytext("y = ['a', 'b', 'c']")
        self.trytext("rep_x = repr(x['a'])")
        self.trytext("rep_y = repr(y)")
        self.trytext("print(rep_y , rep_x)")

        self.isValue("rep_x", "1")
        self.isValue("rep_y", "['a', 'b', 'c']")

    def test_cmp(self):
        '''numeric comparisons'''
        self.isTrue("3 == 3")
        self.isTrue("3.0 == 3")
        self.isTrue("3.0 == 3.0")
        self.isTrue("3 != 4")
        self.isTrue("3.0 != 4")
        self.isTrue("3 >= 1")
        self.isTrue("3 >= 3")
        self.isTrue("3 <= 3")
        self.isTrue("3 <= 5")
        self.isTrue("3 < 5")
        self.isTrue("5 > 3")

        self.isFalse("3 == 4")
        self.isFalse("3 > 5")
        self.isFalse("5 < 3")

    def test_bool(self):
        '''boolean logic'''

        self.trytext('''
yes = True
no = False
nottrue = False
a = arange(7)''')

        self.isTrue("yes")
        self.isFalse("no")
        self.isFalse("nottrue")
        self.isFalse("yes and no or nottrue")
        self.isFalse("yes and (no or nottrue)")
        self.isFalse("(yes and no) or nottrue")
        self.isTrue("yes or no and nottrue")
        self.isTrue("yes or (no and nottrue)")
        self.isFalse("(yes or no) and nottrue")
        self.isTrue("yes or not no")
        self.isTrue("(yes or no)")
        self.isFalse("not (yes or yes)")
        self.isFalse("not (yes or no)")
        self.isFalse("not (no or yes)")
        self.isTrue("not no or yes")
        self.isFalse("not yes")
        self.isTrue("not no")

    def test_bool_coerce(self):
        '''coercion to boolean'''

        self.isTrue("1")
        self.isFalse("0")

        self.isTrue("'1'")
        self.isFalse("''")

        self.isTrue("[1]")
        self.isFalse("[]")

        self.isTrue("(1)")
        self.isTrue("(0,)")
        self.isFalse("()")

        self.isTrue("dict(y=1)")
        self.isFalse("{}")

    def test_assignment(self):
        '''variables assignment'''
        self.trytext('n = 5')
        self.isValue("n",  5)
        self.trytext('s1 = "a string"')
        self.isValue("s1",  "a string")
        self.trytext('b = (1,2,3)')
        self.isValue("b",  (1,2,3))
        self.trytext('a = 1.*arange(10)')
        self.isValue("a", np.arange(10) )
        self.trytext('a[1:5] = 1 + 0.5 * arange(4)')
        self.isNear("a", np.array([ 0. ,  1. ,  1.5,  2. ,  2.5,  5. ,  6. ,  7. ,  8. ,  9. ]))

    def test_names(self):
        '''names test'''
        self.trytext('nx = 1')
        self.trytext('nx1 = 1')

    def test_syntaxerrors_1(self):
        '''assignment syntax errors test'''
        for expr in ('class = 1', 'for = 1', 'if = 1', 'raise = 1',
                     '1x = 1', '1.x = 1', '1_x = 1'):
            failed, errtype, errmsg = False, None, None
            out, err = self.trytext(expr)
            if len(err) > 0:
                errtype, errmsg = err[0].get_error()
                failed = True
            self.assertTrue(failed)
            self.assertTrue(errtype == 'SyntaxError')
            #self.assertTrue(errmsg.startswith('invalid syntax'))

    def test_unsupportednodes(self):
        '''unsupported nodes'''

        for expr in ('f = lambda x: x*x', 'yield 10'):
            failed, errtype, errmsg = False, None, None
            out, err = self.trytext(expr)
            if len(err) > 0:
                errtype, errmsg = err[0].get_error()
                failed = True

            self.assertTrue(failed)
            self.assertTrue(errtype == 'NotImplementedError')

    def test_syntaxerrors_2(self):
        '''syntax errors test'''
        for expr in ('x = (1/*)', 'x = 1.A', 'x = A.2'):

            failed, errtype, errmsg = False, None, None
            out, err = self.trytext(expr)
            if len(err) > 0:
                errtype, errmsg = err[0].get_error()
                failed = True

            self.assertTrue(failed)
            self.assertTrue(errtype == 'SyntaxError')
            #self.assertTrue(errmsg.startswith('invalid syntax'))

    def test_runtimeerrors_1(self):
        '''runtime errors test'''
        self.trytext("zero = 0")
        self.trytext("astr ='a string'")
        self.trytext("atup = ('a', 'b', 11021)")
        self.trytext("arr  = arange(20)")
        for expr, errname in (('x = 1/zero', 'ZeroDivisionError'),
                              ('x = zero + nonexistent', 'NameError'),
                              ('x = zero + astr', 'TypeError'),
                              ('x = zero()', 'TypeError'),
                              ('x = astr * atup', 'TypeError'),
                              ('x = arr.shapx', 'AttributeError'),
                              ('arr.shapx = 4', 'AttributeError'),
                              ('del arr.shapx', 'LookupError')):
            failed, errtype, errmsg = False, None, None

            out, err = self.trytext(expr)
            if len(err) > 0:
                errtype, errmsg = err[0].get_error()
                failed = True
            self.assertTrue(failed)
            self.assertTrue(errtype == errname)
            #self.assertTrue(errmsg.startswith('invalid syntax'))

    def test_ndarrays(self):
        '''simple ndarrays'''
        self.trytext('n = array([11, 10, 9])')
        self.isTrue("isinstance(n, ndarray)")
        self.isTrue("len(n) == 3")
        self.isValue("n", np.array([11, 10, 9]))
        self.trytext('n = arange(20).reshape(5, 4)')
        self.isTrue("isinstance(n, ndarray)")
        self.isTrue("n.shape == (5, 4)")
        self.trytext("myx = n.shape")
        self.trytext("n.shape = (4, 5)")
        self.isTrue("n.shape == (4, 5)")

        # self.trytext("del = n.shape")
        self.trytext("a = arange(20)")
        self.trytext("gg = a[1:13:3]")
        self.isValue('gg', np.array([1, 4, 7, 10]))

        self.trytext("gg[:2] = array([0,2])")
        self.isValue('gg', np.array([0, 2, 7, 10]))
        self.trytext('a, b, c, d = gg')
        self.isValue('c', 7)
        self.isTrue('(a, b, d) == (0, 2, 10)')

    def test_binop(self):
        '''test binary ops'''
        self.trytext('a = 10.0')
        self.trytext('b = 6.0')

        self.isTrue("a+b == 16.0")
        self.isNear("a-b", 4.0)
        self.isTrue("a/(b-1) == 2.0")
        self.isTrue("a*b     == 60.0")

    def test_unaryop(self):
        '''test binary ops'''
        self.trytext('a = -10.0')
        self.trytext('b = -6.0')

        self.isNear("a", -10.0)
        self.isNear("b", -6.0)

    def test_del(self):
        '''test del function'''
        self.trytext('a = -10.0')
        self.trytext('b = -6.0')

        self.assertTrue(self.symtable.has_symbol('a'))
        self.assertTrue(self.symtable.has_symbol('b'))

        self.trytext("del a")
        self.trytext("del b")

        self.assertFalse(self.symtable.has_symbol('a'))
        self.assertFalse(self.symtable.has_symbol('b'))

    def test_math1(self):
        '''builtin math functions'''
        self.trytext('n = sqrt(4)')
        self.isTrue('n == 2')
        self.isNear('sin(pi/2)', 1)
        self.isNear('cos(pi/2)', 0)
        self.isTrue('exp(0) == 1')
        self.isNear('exp(1)', np.e)

    def test_list_comprehension(self):
        "test list comprehension"
        self.trytext('x = [i*i for i in range(4)]')
        self.isValue('x', [0, 1, 4, 9])

        self.trytext('x = [i*i for i in range(6) if i > 1]')
        self.isValue('x', [4, 9, 16, 25])

    def test_ifexp(self):
        "test if expressions"
        self.trytext('x = 2')
        self.trytext('y = 4 if x > 0 else -1')
        self.trytext('z = 4 if x > 3 else -1')
        self.isValue('y', 4)
        self.isValue('z', -1)

    def test_index_assignment(self):
        "test indexing / subscripting on assignment"
        self.trytext('x = arange(10)')
        self.trytext('l = [1,2,3,4,5]')
        self.trytext('l[0] = 0')
        self.trytext('l[3] = -1')
        self.isValue('l', [0,2,3,-1,5])
        self.trytext('l[0:2] = [-1, -2]')
        self.isValue('l', [-1,-2,3,-1,5])

        self.trytext('x[1] = 99')
        self.isValue('x', np.array([0,99,2,3,4,5,6,7,8,9]))
        self.trytext('x[0:2] = [9,-9]')
        self.isValue('x', np.array([9,-9,2,3,4,5,6,7,8,9]))

    def test_reservedwords(self):
        "test reserved words"
        for w in ('and', 'as', 'while', 'raise', 'else',
                  'class', 'del', 'def', 'None', 'True', 'False'):
            out, err = self.trytext("%s= 2" % w)
            self.assertTrue(len(err) > 0)
            errtype, errmsg = err[0].get_error()
            self.assertTrue(errtype=='SyntaxError')

    def test_raise(self):
        "test raise"
        out, err = self.trytext("raise NameError('bob')")
        errtype, errmsg = err[0].get_error()
        errmsgs = errmsg.split('\n')
        self.assertTrue(errtype == 'NameError')


    def test_tryexcept(self):
        "test try/except"
        self.trytext("""
x = 5
try:
    x = x/0
except ZeroDivsionError:
    x = -999
endtry
""")
        self.isValue('x', -999)

        self.trytext("""
x = -1
try:
    x = x/0
except ZeroDivsionError:
    pass
endtry
""")
        self.isValue('x', -1)


    def test_astdump(self):
        "test ast parsing and dumping"
        astnode = self.session._larch.parse('x = 1')
        self.assertTrue(isinstance(astnode, ast.Module))
        self.assertTrue(isinstance(astnode.body[0], ast.Assign))
        self.assertTrue(isinstance(astnode.body[0].targets[0], ast.Name))
        self.assertTrue(isinstance(astnode.body[0].value, ast.Constant))
        self.assertTrue(astnode.body[0].targets[0].id == 'x')
        self.assertTrue(astnode.body[0].value.value == 1)
        dumped = self.session._larch.dump(astnode.body[0])
        self.assertTrue(dumped.startswith('Assign'))

    def test_import(self):
        '''simple import'''
        self.trytext("import numpy")
        self.isTrue("callable(getattr(numpy, 'sqrt'))")
        self.isTrue("callable(getattr(numpy, 'arange'))")
        self.isNear("getattr(numpy, 'pi', 0)", 3.14159, places=5)

    def test_import_as(self):
        '''simple import'''
        self.trytext("import numpy as np")
        self.isTrue("callable(getattr(np, 'sqrt'))")
        self.isTrue("callable(getattr(np, 'arange'))")
        self.isNear("getattr(np, 'pi', 0)", 3.14159, places=5)


if __name__ == '__main__':  # pragma: no cover
    for suite in (TestEval,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
from six.moves.xmlrpc_client import ServerProxy

import time
import json
from larch.utils.jsonutils import decode4js
s = ServerProxy('http://127.0.0.1:4966')

print('Avaialable Methods from XML-RPC server: ', s.system.listMethods())
s.larch('m = 222.3')

s.larch('g = group(x=linspace(0, 10, 11))')
s.larch('g.z = cos(g.x)')

# show and print will be done in server process of course!!!
s.larch('show(g)')

s.larch('print( g.z[3:10])')

print( '== Messages:')
print( s.get_messages())
print( '==')


gx  = decode4js(s.get_data('g.z'))
print( 'm = ', s.get_data('m'))
print( 'x = ', s.get_data('x'))

print('gx = ',  gx, type(gx), gx.dtype)

# could tell server to exit!
# s.exit()



import os
import sqlite3
from base64 import b64encode, b64decode

import numpy as np

from sqlalchemy import MetaData, create_engine, func, text, and_
from sqlalchemy.sql import select
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import SingletonThreadPool

try:
    from pymatgen.io.cif import CifParser
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
    from pymatgen.core import Molecule, IMolecule, IStructure
    from pymatgen.core import __version__ as pmg_version
except:
    CifParser = SpacegroupAnalyzer = None
    Molecule = IMolecule = IStructure = None
    pmg_version = None
    
from larch.utils.physical_constants import ATOM_SYMS, ATOM_NAMES

__version__ = '1'

PMG_CIF_OPTS = dict(occupancy_tolerance=10, site_tolerance=5e-3)


def make_engine(dbname):
    "create engine for sqlite connection"
    return create_engine('sqlite:///%s' % (dbname),
                         poolclass=SingletonThreadPool,
                         connect_args={'check_same_thread': False})

def isAMCSD(dbname):
    """whether a file is a valid AMCSD database

    Args:
        dbname (string): name of AMCSD database file

    Returns:
        bool: is file a valid AMCSD database

    Notes:
      1. must be a sqlite db file, with tables
        'cif', 'elements', 'spacegroup'
    """
    _tables = ('cif', 'elements', 'spacegroups')
    result = False
    try:
        engine = make_engine(dbname)
        meta = MetaData()
        meta.reflect(bind=engine)
        result = all([t in meta.tables for t in _tables])
    except:
        pass
    return result


farray_scale = 4.e6

def encode_farray(dat):
    """encodes a list of fractional coordinate as strings (stricly on (-1,1))
    to an string for saving to db, to be decoded by  decode_farray()
    preserves precision to slightly better than 6 digits
    """
    work = []
    for d in dat:
        if d == '?':
            work.append(2) # out-of-bounds as '?'
        elif d == '.':
            work.append(3) # out-of-bounds as '.'
        else:
            if '(' in d or '(' in d:
                d = d.replace(')', ' : ').replace('(', ' : ')
                d = d.split(':')[0].strip()
            try:
                fval = float(d)
            except ValueError:
                d  = '0'
            work.append(d)
    x = (farray_scale*np.array([float(x) for x in work])).round()
    return b64encode(x.astype(np.int32).tobytes()).decode('ascii')

def decode_farray(dat):
    """decodes a string encoded by encode_farray()
    returns list of string
    """
    arr = np.fromstring(b64decode(dat), dtype=np.int32)/farray_scale
    out = []
    for a in arr:
        if (abs(a-2.0) < 1.e-5):
            out.append('?')
        elif (abs(a-3.0) < 1.e-5):
            out.append('.')
        else:
            out.append(f"{a:f}")
    return out

def put_optarray(dat, attr):
    d = dat.get(attr, '0')
    if d != '0':
        d = encode_farray(d)
    return d

def get_optarray(dat):
    if dat not in (0, '0'):
        dat = decode_farray(dat)
    return dat


schema = (
    '''CREATE TABLE version (id integer primary key, tag text, date text, notes text);''',
    '''CREATE TABLE elements (
        id  integer not null,
        z INTEGER NOT NULL,
        name VARCHAR(40),
        symbol VARCHAR(2) NOT NULL primary key);''',

    '''CREATE TABLE spacegroups (
        id INTEGER primary key,
        hm_notation VARCHAR(16) not null unique,
        symmetry_xyz text NOT NULL,
        category text     );''',

    '''CREATE TABLE minerals (
        id INTEGER not null primary key,
        name text not null unique);''',

    '''CREATE TABLE authors (
        id INTEGER NOT NULL primary key,
        name text unique);''',
    '''CREATE TABLE publications (
        id INTEGER NOT NULL primary key,
        journalname text not null,
        volume text,
        year  integer not null,
        page_first text,
        page_last text);''',

    '''CREATE TABLE publication_authors (
        publication_id INTEGER not null,
        author_id integer not null,
        FOREIGN KEY(publication_id) REFERENCES publications (id),
        FOREIGN KEY(author_id) REFERENCES authors (id));''',

    '''CREATE TABLE cif (
        id integer not null primary key,
        mineral_id INTEGER,
        spacegroup_id INTEGER,
        publication_id INTEGER,
        formula text,
        compound text,
        pub_title text,
        formula_title text,
        a text,
        b text,
        c text,
        alpha text,
        beta text,
        gamma text,
        cell_volume text,
        crystal_density text,
        atoms_sites text,
        atoms_x text,
        atoms_y text,
        atoms_z text,
        atoms_occupancy text,
        atoms_u_iso text,
        atoms_aniso_label text,
        atoms_aniso_u11 text,
        atoms_aniso_u22 text,
        atoms_aniso_u33 text,
        atoms_aniso_u12 text,
        atoms_aniso_u13 text,
        atoms_aniso_u23 text,
        qdat text,
        amcsd_url text,
        FOREIGN KEY(spacegroup_id) REFERENCES spacegroups (id),
        FOREIGN KEY(mineral_id) REFERENCES minerals (id),
        FOREIGN KEY(publication_id) REFERENCES publications (id));''',

    '''CREATE TABLE cif_elements (
        cif_id text not null,
        element VARCHAR(2) not null);''',
    )


def create_amcsd(dbname='test.db'):
    if os.path.exists(dbname):
        os.unlink(dbname)

    conn = sqlite3.connect(dbname)
    cursor = conn.cursor()
    for s in schema:
        cursor.execute(s)

    cursor.execute('insert into version values (?,?,?,?)',
                   ('0', 'in progress', 'today', 'in progress'))

    atz, i = 0, 0
    for sym, name in zip(ATOM_SYMS, ATOM_NAMES):
        i += 1
        atz += 1
        if sym == 'D':
            atz = 1
        cursor.execute('insert into elements values (?,?,?,?)', (i, atz, sym, name))



#!/usr/bin/env python
""" Tests of Larch Scripts  """
import unittest
from pathlib import Path
import time
import ast
import numpy as np
from sys import version_info

from utils import TestCase
from larch import Interpreter

base_dir = Path(__file__).parent.parent.resolve()

class TestScripts(TestCase):
    '''testing of asteval'''
    def test01_basic(self):
        self.runscript('a.lar', dirname=base_dir / 'tests' / 'larch_scripts')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("n < 10")
        self.isTrue("n >  5")
        self.isTrue("x >  3")

    def test02_autobk(self):
        self.runscript('doc_autobk1.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("cu.e0 > 8950.0")
        self.isTrue("len(cu.k) > 200")
        self.isTrue("max(abs(cu.chi)) < 2.0")

    def test03_autobk2(self):
        self.runscript('doc_autobk2.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("dat.e0 > 10000.0")
        self.isTrue("len(dat.k) > 200")

    def test04_autobk_clamp(self):
        self.runscript('doc_autobk3.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("dat.e0 > 11000.0")
        self.isTrue("len(dat.k) > 200")

    def test05_autobk_with_std(self):
        self.runscript('doc_autobk4.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("cu2.e0 > 8950.0")
        self.isTrue("len(cu2.k) > 200")
        self.isTrue("max(abs(cu2.chi)) < 2.0")

    def test06_ftwin1(self):
        self.runscript('doc_ftwin1.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(hann_win1) == 401")
        self.isTrue("hann_win3.sum() > 50.0")

        self.runscript('doc_ftwin2.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(kai_win1) == 401")
        self.isTrue("kai_win1.sum() > 20.0")

    def test07_xafsft1(self):
        self.runscript('doc_xafsft1.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(d2.k) > 200")
        self.isTrue("len(d2.kwin) > 200")
        self.isTrue("d1.chir_mag.sum() > 30")
        self.isTrue("where(d1.chir_mag>1)[0][0] > 60")

    def test08_xafsft2(self):
        self.runscript('doc_xafsft2.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(d3.k) > 200")
        self.isTrue("len(d3.kwin) > 200")
        self.isTrue("len(d4.k) > 200")
        self.isTrue("len(d4.kwin) > 200")
        self.isTrue("len(d1.r) > 100")
        self.isTrue("len(d1.chir_mag) > 100")
        self.isTrue("len(d3.r) > 100")
        self.isTrue("len(d3.chir_mag) > 100")
        self.isTrue("len(d4.r) > 100")
        self.isTrue("len(d4.chir_mag) > 100")
        self.isTrue("len(d4.chir_re) > 100")
        self.isTrue("len(d4.chir_im) > 100")


    def test09_xafsft3(self):
        self.runscript('doc_xafsft3.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(dat.k) > 200")
        self.isTrue("len(dat.kwin) > 200")

    def test10_xafsft3(self):
        self.runscript('doc_xafsft4.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("len(dat.r) > 200")
        self.isTrue("len(dat.rwin) > 200")
        self.isTrue("len(dat.q) > 200")
        self.isTrue("len(dat.chiq_re) > 200")

    def test11_wavelet1(self):
        self.runscript('wavelet_example.lar', dirname=base_dir / 'examples' / 'xafs')
        assert(len(self.session.get_errors()) == 0)
        self.isTrue("f.wcauchy_im.shape == (326, 318)")
        self.isTrue("f.wcauchy_mag.sum() > 300")

    def test12_feffit_kws(self):
        self.runscript('test_epsk_kws.lar', dirname=base_dir / 'examples' / 'feffit')
        assert(len(self.session.get_errors()) == 0)
        out = self.session.run('out')
        for row in out:
            amp = row[5]
            amp_err = row[5]
            delr= row[7]
            self.assertTrue(amp > 0.5)
            self.assertTrue(amp < 2.0)
            self.assertTrue(amp_err > 0.0)
            self.assertTrue(amp_err < 2.0)
            self.assertTrue(abs(delr) < 0.1)

    def test13_feffit1(self):
        self.runscript('doc_feffit1.lar', dirname=base_dir / 'examples' / 'feffit')
        assert(len(self.session.get_errors()) == 0)

        self.isTrue('out.nfev > 20')
        self.isTrue('out.nfev < 100')
        self.isTrue('out.chi_square > 0.2')
        self.isTrue('out.chi_square < 2000')
        self.isNear('pars.amp.value',     1.0, places=1)
        self.isNear('pars.del_e0.value',    0, places=1)
        self.isNear('pars.del_r.value',     0, places=1)
        self.isNear('out.paramgroup.amp.value',     0.93, places=1)
        self.isNear('out.paramgroup.del_e0.value',  4.364, places=1)
        self.isNear('out.paramgroup.del_e0.stderr', 0.52, places=1)
        self.isNear('out.paramgroup.del_r.value',  -0.006, places=3)
        self.isNear('out.paramgroup.sig2.value',    0.0087, places=3)

    def test14_feffdat3(self):
        self.runscript('doc_feffdat3.lar', dirname=base_dir / 'examples' / 'feffit')
        assert(len(self.session.get_errors()) == 0)

        self.isNear('path1.degen',  6.0, places=1)
        self.isNear('path1.e0', -0.5, places=1)
        self.isNear('path1.reff',  2.1387, places=2)
        self.isNear('path1.rmass',  12.436, places=2)
        self.isNear('path1.s02',     0.9, places=2)
        self.isNear('path1.sigma2',  0.003, places=2)

        self.isTrue('path1.deltar == 0')
        self.isTrue('path1.ei == 0')
        self.isTrue('path1.filename == "feff_feo01.dat"')
        self.isTrue('path1.fourth == 0')
        self.isTrue('path1.nleg == 2')
        self.isTrue('path1.third == 0')
        self.isTrue('path1.geom[0][0] == "Fe"')
        self.isTrue('path1.geom[1][0] == "O"')
        self.isTrue('path1.geom[0][1] == 26')
        self.isTrue('path1.geom[1][1] == 8')


    def test15_feffit2(self):
        self.runscript('doc_feffit2.lar', dirname=base_dir / 'examples' / 'feffit')
        assert(len(self.session.get_errors()) == 0)

        self.isTrue('out.nfev > 50')
        self.isTrue('out.nfev < 200')

        self.isTrue('out.chi_square > 4')
        self.isTrue('out.chi_square < 100')
        self.isTrue('out.rfactor < 0.003')
        self.isTrue('out.aic < 1000')
        self.isNear("out.params['sig2_1'].value", 0.00868, places=2)
        self.isNear("out.params['del_e0'].value",    5.75,    places=1)
        self.isNear("out.params['amp'].value",    0.933,  places=1)

        self.isNear('path1.reff',    2.5478,  places=3)
        self.isNear('path1.rmass',  31.773,   places=2)


    def test16_feffit3(self):
        self.runscript('doc_feffit3.lar', dirname=base_dir / 'examples' / 'feffit')
        assert(len(self.session.get_errors()) == 0)

        self.isTrue('out.nfev > 15')
        self.isTrue('out.nfev < 50')

        self.isTrue('out.chi_square > 140')
        self.isTrue('out.chi_square < 10000')

        self.isTrue('out.aic < 1000')
        self.isNear("out.params['theta'].value", 233., places=0)
        self.isNear("out.params['del_e0'].value", 5.3,  places=0)
        self.isNear("out.params['amp'].value", 0.87,  places=0)


    def test17_feffit3extra(self):
        self.runscript('doc_feffit3.lar', dirname=base_dir / 'examples' / 'feffit')
        self.runscript('doc_feffit3_extra.lar', dirname=base_dir / 'examples' / 'feffit')
        assert(len(self.session.get_errors()) == 0)
        self.isNear('_ave', 0.005030, places=4)
        self.isNear('_dlo', 0.000315, places=4)




if __name__ == '__main__':  # pragma: no cover
    for suite in (TestScripts,):
        suite = unittest.TestLoader().loadTestsFromTestCase(suite)
        unittest.TextTestRunner(verbosity=13).run(suite)



from .nexus import NexusSingleXasDataSource
from .nexus import EsrfSingleXasDataSource
from .nexus import SoleilSingleXasDataSource
from .spec import SpecSingleXasDataSource

_SOURCE_TYPES = {
    "nexus": NexusSingleXasDataSource,
    "esrf": EsrfSingleXasDataSource,
    "soleil": SoleilSingleXasDataSource,
    "spec": SpecSingleXasDataSource,
}


def get_source_type(name):
    return _SOURCE_TYPES[name]



"""
Adds subfigure functionality
"""

from docutils import nodes
import docutils.parsers.rst.directives as directives
from docutils.parsers.rst import Directive
from sphinx import addnodes

class subfig(nodes.General, nodes.Element):
    pass

def skip_visit(self, node):
    raise nodes.SkipNode

def visit_subfig_tex(self, node):
    self.__body = self.body
    self.body = []

def depart_subfig_tex(self, node):
    figoutput = ''.join(self.body)
    figoutput = figoutput.replace('[tbp]', '[t]{%s\\linewidth}' % node['width'])
    figoutput = figoutput.replace('figure', 'subfigure')
    self.body = self.__body
    self.body.append(figoutput)

def visit_subfig_html(self, node):
    self.__body = self.body
    self.body = []

def depart_subfig_html(self, node):
    figoutput = ''.join(self.body)
    figoutput = figoutput.replace('class="figure', 'style="width: %g%%" class="subfigure' % (float(node['width']) * 100))
    self.body = self.__body
    self.body.append(figoutput)

class subfigstart(nodes.General, nodes.Element):
    pass

def visit_subfigstart_tex(self, node):
    self.body.append('\n\\begin{figure}\n\\centering\n\\capstart\n')

def depart_subfigstart_tex(self, node):
    pass

def visit_subfigstart_html(self, node):
    atts = {'class': 'figure compound align-center'}
    self.body.append(self.starttag(node['subfigend'], 'div', **atts))

def depart_subfigstart_html(self, node):
    pass

class subfigend(nodes.General, nodes.Element):
    pass

def visit_subfigend_tex(self, node):
    pass

def depart_subfigend_tex(self, node):
    self.body.append('\n\n\\end{figure}\n\n')

def visit_subfigend_html(self, node):
    pass

def depart_subfigend_html(self, node):
    self.body.append('</div>')

class SubFigEndDirective(Directive):
    has_content = True
    optional_arguments = 3
    final_argument_whitespace = True

    option_spec = {'label': directives.uri,
                   'alt': directives.unchanged,
                   'width': directives.unchanged_required}

    def run(self):
        label = self.options.get('label', None)
        width = self.options.get('width', None)
        alt = self.options.get('alt', None)

        node = subfigend('', ids=[label] if label is not None else [])

        if width is not None:
            node['width'] = width
        if alt is not None:
            node['alt'] = alt

        if self.content:
            anon = nodes.Element()
            self.state.nested_parse(self.content, self.content_offset, anon)
            first_node = anon[0]
            if isinstance(first_node, nodes.paragraph):
                caption = nodes.caption(first_node.rawsource, '',
                                        *first_node.children)
                node += caption

        if label is not None:
            targetnode = nodes.target('', '', ids=[label])
            node.append(targetnode)

        return [node]

class SubFigStartDirective(Directive):
    has_content = False
    optional_arguments = 0

    def run(self):
        node = subfigstart()
        return [node]

def doctree_read(app, doctree):
    secnums = app.builder.env.toc_secnumbers
    for node in doctree.traverse(subfigstart):
        parentloc = node.parent.children.index(node)

        subfigendloc = parentloc
        while subfigendloc < len(node.parent.children):
            n = node.parent.children[subfigendloc]
            if isinstance(n, subfigend):
                break
            subfigendloc += 1

        if subfigendloc == len(node.parent.children):
            return

        between_nodes = node.parent.children[parentloc:subfigendloc]
        subfigend_node = node.parent.children[subfigendloc]
        node['subfigend'] = subfigend_node
        for i, n in enumerate(between_nodes):
            if isinstance(n, nodes.figure):
                children = [n]
                prevnode = between_nodes[i-1]
                if isinstance(prevnode, nodes.target):
                    node.parent.children.remove(prevnode)
                    children.insert(0, prevnode)
                nodeloc = node.parent.children.index(n)
                node.parent.children[nodeloc] = subfig('', *children)
                node.parent.children[nodeloc]['width'] = subfigend_node['width']
                node.parent.children[nodeloc]['mainfigid'] = subfigend_node['ids'][0]

def setup(app):
    app.add_node(subfigstart,
                 html=(visit_subfigstart_html, depart_subfigstart_html),
                 singlehtml=(visit_subfigstart_html, depart_subfigstart_html),
                 text=(skip_visit, None),
                 latex=(visit_subfigstart_tex, depart_subfigstart_tex))

    app.add_node(subfig,
                 html=(visit_subfig_html, depart_subfig_html),
                 singlehtml=(visit_subfig_html, depart_subfig_html),
                 text=(skip_visit, None),
                 latex=(visit_subfig_tex, depart_subfig_tex))

    app.add_node(subfigend,
                 html=(visit_subfigend_html, depart_subfigend_html),
                 singlehtml=(visit_subfigend_html, depart_subfigend_html),
                 text=(skip_visit, None),
                 latex=(visit_subfigend_tex, depart_subfigend_tex))

    app.add_directive('subfigstart', SubFigStartDirective)
    app.add_directive('subfigend', SubFigEndDirective)

    app.connect('doctree-read', doctree_read)



"""
Adds a new directive called 'figtable' that creates a figure
around a table.
"""

from docutils import nodes
import docutils.parsers.rst.directives as directives
from docutils.parsers.rst import Directive
from sphinx import addnodes

class figtable(nodes.General, nodes.Element):
    pass

def visit_figtable_node(self, node):
    pass

def depart_figtable_node(self, node):
    pass

def visit_figtable_tex(self, node):
    if node['nofig']:
        self.body.append('\n\n\\begin{table}\n\\capstart\n\\begin{center}\n')
    else:
        self.body.append('\n\n\\begin{figure}[tbp]\n\\capstart\n\\begin{center}\n')

def depart_figtable_tex(self, node):
    if node['nofig']:
        self.body.append('\n\\end{center}\n\\end{table}\n')
    else:
        self.body.append('\n\\end{center}\n\\end{figure}\n')

def visit_figtable_html(self, node):
    atts = {'class': 'figure align-center'}
    self.body.append(self.starttag(node, 'div', **atts) + '<center>')

def depart_figtable_html(self, node):
    self.body.append('</center></div>')

class FigTableDirective(Directive):

    has_content = True
    optional_arguments = 5
    final_argument_whitespace = True

    option_spec = {'label': directives.uri,
                   'spec': directives.unchanged,
                   'caption': directives.unchanged,
                   'alt': directives.unchanged,
                   'nofig': directives.flag}

    def run(self):
        label = self.options.get('label', None)
        spec = self.options.get('spec', None)
        caption = self.options.get('caption', None)
        alt = self.options.get('alt', None)
        nofig = 'nofig' in self.options

        figtable_node = figtable('', ids=[label] if label is not None else [])
        figtable_node['nofig'] = nofig

        if spec is not None:
            table_spec_node = addnodes.tabular_col_spec()
            table_spec_node['spec'] = spec
            figtable_node.append(table_spec_node)

        node = nodes.Element()
        self.state.nested_parse(self.content, self.content_offset, node)
        tablenode = node[0]
        if alt is not None:
            tablenode['alt'] = alt
        figtable_node.append(tablenode)

        if caption is not None:
            caption_node = nodes.caption('', '', nodes.Text(caption))
            figtable_node.append(caption_node)

        if label is not None:
            targetnode = nodes.target('', '', ids=[label])
            figtable_node.append(targetnode)

        return [figtable_node]

def setup(app):
    app.add_node(figtable,
                 html=(visit_figtable_html, depart_figtable_html),
                 singlehtml=(visit_figtable_html, depart_figtable_html),
                 latex=(visit_figtable_tex, depart_figtable_tex),
                 text=(visit_figtable_node, depart_figtable_node))

    app.add_directive('figtable', FigTableDirective)



#!/usr/bin/env python
"""
   feffit sums Feff paths to match xafs data
"""

import sys, os
import numpy as np
from numpy import arange, interp, pi, zeros, sqrt
from numpy.fft import fft, ifft

from scipy.optimize import leastsq as scipy_leastsq

import larch
from larch.larchlib import Parameter, isParameter, plugin_path
from larch.utils import OrderedDict
from larch.symboltable import isgroup


sys.path.insert(0, plugin_path('std'))
sys.path.insert(0, plugin_path('fitter'))
sys.path.insert(0, plugin_path('xafs'))

from mathutils import index_of, realimag

from minimizer import Minimizer
from xafsft import xafsft, xafsft_prep, xafsft_fast, xafsift, ftwindow
from feffdat import FeffPathGroup, _ff2chi

class FilterGroup(larch.Group):
    def __init__(self, kmin=0, kmax=20, kw=2, dk=1, dk2=None, window='kaiser',
                 rmin = 0, rmax=10, dr=0, rwindow='kaiser', kweight=None,
                 nfft=2048, kstep=0.05, fitspace='r', _larch=None, **kws):
        larch.Group.__init__(self)

        self._kmin = kmin
        self._kmax = kmax
        self._kw = kw
        if kweight is not None:
            self._kw = kweight
        self._dk = dk
        self._dk2 = dk2
        self._window = window
        self._rmin = rmin
        self._rmax = rmax
        self._dr = dr
        self._rwindow = rwindow
        self._kstep = kstep

        self.nfft = nfft
        self.fitspace = fitspace
        self._larch = _larch
        self.rstep = pi/(self.kstep*self.nfft)

        self.kwin_array = None
        self.rwin_array = None

    ## these should all be properties that, when set, erase the
    ## corresponding window function, so that the window will be
    ## recalculated the next time it is needed

    @property
    def kmin(self): return self._kmin

    @property
    def kmax(self): return self._kmax

    @property
    def kw(self): return self._kw

    @property
    def dk(self): return self._dk

    @property
    def dk2(self): return self._dk2

    @property
    def window(self): return self._window

    @property
    def rmin(self): return self._rmin

    @property
    def rmax(self): return self._rmax

    @property
    def rw(self): return self._rw

    @property
    def dr(self): return self._dr

    @property
    def rwindow(self): return self._rwindow


    @property
    def kstep(self): return self._kstep

    @kmin.setter
    def kmin(self, val):
        self.kwin_array = None
        self._kmin = val

    @kmax.setter
    def kmax(self, val):
        self.kwin_array = None
        self._kmax = val

    @kw.setter
    def kw(self, val):
        self.kwin_array = None
        self._kw = val

    def fft(self, k, chi):
        if self.kwin_array is None:
            self.kwin_array = ftwindow(k, xmin=self.kmin, xmax=self.kmax,
                                       dx=self.dk, dx2=self.dk2,
                                       window=self.window)
        cchi = zeros(self.nfft, dtype='complex128')
        cchi[0:len(chi)] = chi
        out = fft(cchi * self.kwin_array * k**self.kw)
        return self.kstep*sqrt(pi) * out[:self.nfft/2]

    def ifft(self, r, chir):
        if self.rwin_array is None:
            self.rwin_array = ftwindow(r, xmin=self.rmin, xmax=self.rmax,
                                       dx=self.dr, dx2=self.dr2,
                                       window=self.rwindow)

        cchir = zeros(self.nfft, dtype='complex128')
        cchir[0:len(chir)] = chir
        out = ifft(cchir * self.rwin_array * r**self.rw)
        return sqrt(pi)/(2*self.kstep) * out[:self.nfft/2]

    def apply(self, k, chi, **kws):
        """apply filter"""

        for key, val in kws:
            if key == 'kweight': key = 'kw'
            setattr(self, key, val)

        if self.fitspace == 'k':
            return chi * k**self.kw
        elif self.fitspace in ('r', 'q'):

            k_ = self.kstep * arange(self.nfft, dtype='float64')
            r_ = self.rstep * arange(self.nfft, dtype='float64')
            chir = self.fft(k_, chi)
            if self.fitspace == 'r':
                irmin = index_of(r_, self.rmin)
                irmax = min(self.nfft/2,  1 + index_of(r_, self.rmax))
                return realimag(chir[irmin:irmax])
            else:
                chiq = self.ifft(r_, chir)
                iqmin = index_of(k_, self.kmin)
                iqmax = min(self.nfft/2,  1 + index(k_, self.kmax))
                return realimag(chiq[ikmin:ikmax])

class FeffitDataSet(larch.Group):
    def __init__(self, data=None, pathlist=None, filter=None, _larch=None):
        self._larch = _larch
        larch.Group.__init__(self,  **kws)
        self.pathlist = pathlist

        self.datagroup = data
        if filter is None:
            filter = FilterGroup()
        self.filter = filter
        self.model = Group()

    def calc_model(self):
        """calculate model spectra from pathlist -- essentially ff2chi"""
        _ff2chi(self.pathlist, _larch =_larch, group=self.model)

    def apply_transform(self):
        if not isinstance(self.filter, FilterGroup):
            self.filter.apply(self.datagroup.k, self.datagroup.chi)


class Feffit(Minimizer):
    def __init__(self, pathlist, params, _larch=None, **kws):
        Minimizer.__init__(self, self._feffit, params, _larch=_larch, **kws)

    def _feffit(self, x):
        """residual function for feffit"""

def feffit_fit(_larch=None):
    print( 'not yet!')


def feffit_dataset(data=None, pathlist=None, filter=None, _larch=None):
    return FeffitDataSet(data=data, pathlist=pathlist, filter=filter, _larch=_larch)


def feffit_filter(_larch=None, **kws):
    return FilterGroup(_larch=_larch, **kws)

def registerLarchPlugin():
    print( '====== test FEFFIT ======')
    return ('_xafs', {'feffit_fit': feffit_fit,
                      'feffit_dataset': feffit_dataset,
                      'feffit_filter': feffit_filter,


                      })
